# -*- coding: utf-8 -*-

"""Fixes wrongly preprocessed lists of ints in the preprocessed pandas dataframes.

Run with:
python -m src.stonkgs.data.fix_broken_pretraining_dataset
"""
import os
import pandas as pd
from tqdm import tqdm

from stonkgs.constants import (
    PRETRAINING_DIR,
)


def convert_tsv_with_str_list(
    file_name: str,
    path: str = PRETRAINING_DIR,
):
    """Converts strings that look like lists into actual lists of ints and saves a pickle of the repaired dataframe."""
    converter = lambda x: [int(y) for y in x.strip("[]").split(", ")]
    converter_dict = {
        col: converter
        for col in [
            "input_ids",
            "attention_mask",
            "token_type_ids",
            "masked_lm_labels",
            "ent_masked_lm_labels",
        ]
    }
    chunks = []

    # Process dataframe in bits with a progress bar
    for chunk in tqdm(
        pd.read_csv(
            os.path.join(path, file_name + ".tsv"),
            sep="\t",
            chunksize=10000,
            converters=converter_dict,
        ),
        total=10,  # Hard coded total number
    ):
        chunks.append(chunk)
    complete_df = pd.concat(chunks, axis=0)

    # Pickle the complete dataframe
    complete_df.to_pickle(os.path.join(PRETRAINING_DIR, file_name + ".pkl"))


if __name__ == "__main__":
    # Fixing the positive dataset
    convert_tsv_with_str_list("pretraining_preprocessed_dummy")



# -*- coding: utf-8 -*-

"""Prepares the pre-training data for STonKGs.

Run with:
python -m src.stonkgs.data.indra_for_pretraining
"""

import logging
import os
import random
from typing import List, Optional

import numpy as np
import pandas as pd
from transformers import BertTokenizer, BertTokenizerFast
from tqdm import tqdm

from stonkgs.constants import (
    EMBEDDINGS_PATH,
    NLP_MODEL_TYPE,
    PRETRAINING_DIR,
    PRETRAINING_PATH,
    RANDOM_WALKS_PATH,
    VOCAB_FILE,
)
from stonkgs.models.kg_baseline_model import prepare_df

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def replace_mlm_tokens(
    tokens: List[int],
    vocab_len: int,
    mask_id: int = 103,
    masked_tokens_percentage: float = 0.15,
    unmasked_label_id: int = -100,
):
    """Applies masking to the given sequence with numeric indices and returns the manipulated sequence + labels."""
    # This code is taken from: https://d2l.ai/chapter_natural-language-processing-pretraining/
    # bert-dataset.html#defining-helper-functions-for-pretraining-tasks

    # Make a new copy of tokens for the input of a masked language model,
    # where the input may contain replaced '<mask>' or random tokens
    mlm_input_tokens = [token for token in tokens]

    # The unmasked_label_id indicates that these are NOT the labels that need to be predicted in MLM
    # For MLM, the model expects a tensor of dimension(batch_size, seq_length) with each value
    # corresponding to the expected label of each individual token: the labels being the token ID
    # for the masked token, and values to be ignored for the rest (usually -100).
    mlm_labels = [unmasked_label_id] * len(mlm_input_tokens)

    # Shuffle for getting 15% random tokens for prediction in the MLM task
    candidate_pred_positions = random.sample(
        range(len(mlm_input_tokens)),
        int(len(mlm_input_tokens) * masked_tokens_percentage),
    )

    for mlm_pred_position in candidate_pred_positions:
        # 80% of the time: replace the word with the '<mask>' token
        if random.random() < 0.8:
            masked_token = mask_id
        else:
            # 10% of the time: keep the word unchanged
            if random.random() < 0.5:
                masked_token = tokens[mlm_pred_position]
            # 10% of the time: replace the word with a random word
            else:
                masked_token = random.randint(0, vocab_len - 1)

        # Replace accordingly in the original sequence
        mlm_input_tokens[mlm_pred_position] = masked_token
        # Use the original token label for the list of masked labels
        mlm_labels[mlm_pred_position] = tokens[mlm_pred_position]

    return mlm_input_tokens, mlm_labels


def _add_negative_nsp_samples(
    processed_df: pd.DataFrame,
    nsp_negative_proportion: float = 0.25,
):
    """Generates non-matching text-entity pairs (negative NSP samples)."""
    negative_samples = []

    # Get the length of half a sequence
    half_length = len(processed_df.iloc[0]["input_ids"]) // 2

    # First get the indices that serve as the basis for the negative examples
    negative_sample_idx = random.sample(
        range(len(processed_df)),
        int(len(processed_df) * nsp_negative_proportion),
    )

    # For each negative sample, get a random index that is not the current one for creating negative samples
    # (doesn't exclude the very very small chance of getting the same index as a partner
    negative_sample_idx_partner = random.sample(
        range(len(processed_df)),
        int(len(processed_df) * nsp_negative_proportion),
    )

    for i, j in tqdm(
        zip(negative_sample_idx, negative_sample_idx_partner),
        total=int(nsp_negative_proportion * len(processed_df)),
        desc="Generating negative samples",
    ):
        # Get the features from i
        text_features = processed_df.iloc[i]
        # Get the features from j
        entity_features = processed_df.iloc[j]
        # 1. Replace the second half of the input sequence
        # 2. Replace the entity mask labels
        # 3. Replace the NSP label
        new_entry = {
            "input_ids": text_features["input_ids"][:half_length]
            + entity_features["input_ids"][half_length:],
            "attention_mask": text_features["attention_mask"],
            "token_type_ids": text_features["token_type_ids"],
            "masked_lm_labels": text_features["masked_lm_labels"],
            "ent_masked_lm_labels": entity_features["ent_masked_lm_labels"],
            "next_sentence_labels": 1,  # 1 indicates it's a corrupted training sample
        }
        negative_samples.append(new_entry)

    negative_samples_df = pd.DataFrame(negative_samples)

    return negative_samples_df


def indra_to_pretraining_df(
    embedding_name_to_vector_path: str = EMBEDDINGS_PATH,
    embedding_name_to_random_walk_path: str = RANDOM_WALKS_PATH,
    pre_training_path: str = PRETRAINING_PATH,
    nlp_model_type: str = NLP_MODEL_TYPE,
    nsp_negative_proportion: float = 0.25,
    sep_id: int = 102,
):
    """Preprocesses the INDRA statements from the pre-training file so that it contains all the necessary attributes."""
    # Load the KG embedding dict to convert the names to numeric indices
    kg_embed_dict = prepare_df(embedding_name_to_vector_path)
    kg_name_to_idx = {key: i for i, key in enumerate(kg_embed_dict.keys())}

    # Load the random walks for each node
    random_walk_dict = prepare_df(embedding_name_to_random_walk_path)
    # Assert that embeddings and random walks are generated based on the same dataset
    assert len(kg_embed_dict) == len(
        random_walk_dict
    ), "Embeddings and random walks must cover the same entities"

    # Log the number of entities
    logger.info(f"There are {len(kg_embed_dict)} many entities in the pre-trained KG")

    # Convert random walk sequences to list of numeric indices
    random_walk_idx_dict = {
        k: [kg_name_to_idx[node] for node in v] for k, v in random_walk_dict.items()
    }

    # Load the pre-training dataframe
    pretraining_df = pd.read_csv(pre_training_path, sep="\t")

    # Check if all pretraining entities are covered by the embedding dict
    number_of_pre_training_nodes = len(
        set(pretraining_df["source"]).union(set(pretraining_df["target"]))
    )
    if number_of_pre_training_nodes > len(kg_embed_dict):
        logger.warning(
            "The learned KG embeddings do not cover all of the nodes in the pre-training data"
        )
        return

    # Get the length of the text or entity embedding sequences (2 random walks + 2 = entity embedding sequence length)
    half_length = len(next(iter(random_walk_idx_dict.values()))) * 2 + 2

    # Initialize a FAST tokenizer if it's the default one (BioBERT)
    if nlp_model_type == NLP_MODEL_TYPE:
        # Initialize the fast tokenizer for getting the text token ids
        tokenizer = BertTokenizerFast(vocab_file=VOCAB_FILE)
    else:
        # Initialize a slow tokenizer used for getting the text token ids
        tokenizer = BertTokenizer.from_pretrained(nlp_model_type)

    # Initialize the preprocessed data
    pre_training_preprocessed = []

    # Log progress with a progress bar
    for idx, row in tqdm(
        pretraining_df.iterrows(),
        total=pretraining_df.shape[0],
        desc="Generating positive samples",
    ):
        # 1. "Token type IDs": 0 for text tokens, 1 for entity tokens
        token_type_ids = [0] * half_length + [1] * half_length

        # 2. Tokenization for getting the input ids and attention masks for the text
        # Use encode_plus to also get the attention mask ("padding" mask)
        encoded_text = tokenizer.encode_plus(
            row["evidence"],
            padding="max_length",
            truncation=True,
            max_length=half_length,
        )
        text_token_ids = encoded_text["input_ids"]
        text_attention_mask = encoded_text["attention_mask"]

        # 3. Get the random walks sequence and the node indices, add the SEP (usually with id=102) in between
        random_walks = (
            random_walk_idx_dict[row["source"]]
            + [sep_id]
            + random_walk_idx_dict[row["target"]]
            + [sep_id]
        )

        # 4. Total attention mask (attention mask is all 1 for the entity sequence)
        attention_mask = text_attention_mask + [1] * half_length

        # Apply the masking strategy to the text tokens and get the text MLM labels
        masked_lm_token_ids, masked_lm_labels = replace_mlm_tokens(
            tokens=text_token_ids,
            vocab_len=len(tokenizer.vocab),
        )
        # Apply the masking strategy to the entity tokens and get the entity (E)LM labels
        # Use the same mask_id as in the NLP model (handled appropriately by STonKGs later on)
        ent_masked_lm_token_ids, ent_masked_lm_labels = replace_mlm_tokens(
            tokens=random_walks,
            vocab_len=len(kg_embed_dict),
        )

        input_ids = masked_lm_token_ids + ent_masked_lm_token_ids

        # Add all the features to the preprocessed data
        pre_training_preprocessed.append(
            {
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "token_type_ids": token_type_ids,
                "masked_lm_labels": masked_lm_labels,
                "ent_masked_lm_labels": ent_masked_lm_labels,
                "next_sentence_labels": 0,  # 0 indicates the random walks belong to the text evidence
            }
        )

    # Put the preprocessed data into a dataframe
    pre_training_preprocessed_df = pd.DataFrame(pre_training_preprocessed)

    logger.info("Finished generating positive training examples")

    # Save the positive examples
    pre_training_preprocessed_df.to_csv(
        os.path.join(PRETRAINING_DIR, "pretraining_preprocessed_positive.tsv"),
        sep="\t",
        index=False,
    )
    # Pickle it, too (easier for reading in the lists in the pandas dataframe)
    pre_training_preprocessed_df.to_pickle(
        os.path.join(PRETRAINING_DIR, "pretraining_preprocessed_positive.pkl"),
    )

    """
    # Use this in case the script crashes during the generation of negative samples
    # Load the positive examples
    pre_training_preprocessed_df = pd.read_pickle(
        os.path.join(PRETRAINING_DIR, 'pretraining_preprocessed_positive.pkl'),
    )
    """

    # Generate the negative NSP training samples
    pre_training_negative_samples = _add_negative_nsp_samples(
        pre_training_preprocessed_df,
        nsp_negative_proportion=nsp_negative_proportion,
    )

    # And append them to the original data
    pre_training_preprocessed_df = pre_training_preprocessed_df.append(
        pre_training_negative_samples
    ).reset_index(drop=True)

    logger.info("Finished generating negative training examples")

    # Shuffle the dataframe just to be sure
    pre_training_preprocessed_df_shuffled = pre_training_preprocessed_df.iloc[
        np.random.permutation(pre_training_preprocessed_df.index)
    ].reset_index(drop=True)

    logger.info("Finished shuffling the data")

    # Save the final dataframe
    pre_training_preprocessed_df_shuffled.to_csv(
        os.path.join(PRETRAINING_DIR, "pretraining_preprocessed.tsv"),
        sep="\t",
        index=False,
    )
    # Pickle it, too (easier for reading in the lists in the pandas dataframe)
    pre_training_preprocessed_df_shuffled.to_pickle(
        os.path.join(PRETRAINING_DIR, "pretraining_preprocessed.pkl"),
    )

    logger.info(f"Saved the data under {PRETRAINING_DIR}")

    return pre_training_preprocessed_df


if __name__ == "__main__":
    # Just simple testing
    indra_to_pretraining_df()



# -*- coding: utf-8 -*-

"""Embeddings from the pre-trained STonKGs model."""

import logging
from typing import List, Optional

import pandas as pd
import torch
from tqdm import tqdm
from transformers import BertTokenizer, BertTokenizerFast

from stonkgs.constants import (
    EMBEDDINGS_PATH,
    NLP_MODEL_TYPE,
    RANDOM_WALKS_PATH,
    VOCAB_FILE,
)
from stonkgs.data.indra_for_pretraining import prepare_df, replace_mlm_tokens
from stonkgs.models.stonkgs_pretraining import STonKGsForPreTraining

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def preprocess_df_for_embeddings(
    df: pd.DataFrame,
    embedding_name_to_vector_path: str = EMBEDDINGS_PATH,
    embedding_name_to_random_walk_path: str = RANDOM_WALKS_PATH,
    nlp_model_type: str = NLP_MODEL_TYPE,
    sep_id: int = 102,
    unk_id: int = 100,
) -> pd.DataFrame:
    """Preprocesses a given pandas dataframe so that it's ready for embedding extraction by STonKGs."""
    # Load the KG embedding dict to convert the names to numeric indices
    kg_embed_dict = prepare_df(embedding_name_to_vector_path)
    kg_name_to_idx = {key: i for i, key in enumerate(kg_embed_dict.keys())}

    # Load the random walks for each node
    random_walk_dict = prepare_df(embedding_name_to_random_walk_path)

    # Log the number of entities
    logger.info(f"There are {len(kg_embed_dict)} many entities in the pre-trained KG")

    # Convert random walk sequences to list of numeric indices
    random_walk_idx_dict = {
        k: [kg_name_to_idx[node] for node in v] for k, v in random_walk_dict.items()
    }

    # Get the length of the text or entity embedding sequences (2 random walks + 2 = entity embedding sequence length)
    half_length = len(next(iter(random_walk_idx_dict.values()))) * 2 + 2

    # Initialize a FAST tokenizer if it's the default one (BioBERT)
    if nlp_model_type == NLP_MODEL_TYPE:
        # Initialize the fast tokenizer for getting the text token ids
        tokenizer = BertTokenizerFast(vocab_file=VOCAB_FILE)
    else:
        # Initialize a slow tokenizer used for getting the text token ids
        tokenizer = BertTokenizer.from_pretrained(nlp_model_type)

    # Initialize the preprocessed data
    pre_training_preprocessed = []

    # Log progress with a progress bar
    for _, row in tqdm(
        df.iterrows(),
        total=df.shape[0],
        desc="Preprocessing the dataset",
    ):
        # 1. "Token type IDs": 0 for text tokens, 1 for entity tokens
        token_type_ids = [0] * half_length + [1] * half_length

        # 2. Tokenization for getting the input ids and attention masks for the text
        # Use encode_plus to also get the attention mask ("padding" mask)
        encoded_text = tokenizer.encode_plus(
            row["evidence"],
            padding="max_length",
            truncation=True,
            max_length=half_length,
        )
        text_token_ids = encoded_text["input_ids"]
        text_attention_mask = encoded_text["attention_mask"]
        random_walk_length = len(next(iter(random_walk_idx_dict.values())))

        # 3. Get the random walks sequence and the node indices, add the SEP (usually with id=102) in between
        # Use a sequence of UNK tokens if the node is not contained in the dictionary of the nodes from pre-training
        random_w_source = (
            random_walk_idx_dict[row["source"]]
            if row["source"] in random_walk_idx_dict.keys()
            else [unk_id] * random_walk_length
        )
        random_w_target = (
            random_walk_idx_dict[row["target"]]
            if row["target"] in random_walk_idx_dict.keys()
            else [unk_id] * random_walk_length
        )
        random_walks = random_w_source + [sep_id] + random_w_target + [sep_id]

        # 4. Total attention mask (attention mask is all 1 for the entity sequence)
        attention_mask = text_attention_mask + [1] * half_length

        # Apply the masking strategy to the text tokens and get the text MLM labels
        masked_lm_token_ids, masked_lm_labels = replace_mlm_tokens(
            tokens=text_token_ids,
            vocab_len=len(tokenizer.vocab),
        )
        # Apply the masking strategy to the entity tokens and get the entity (E)LM labels
        # Use the same mask_id as in the NLP model (handled appropriately by STonKGs later on)
        ent_masked_lm_token_ids, ent_masked_lm_labels = replace_mlm_tokens(
            tokens=random_walks,
            vocab_len=len(kg_embed_dict),
        )

        input_ids = masked_lm_token_ids + ent_masked_lm_token_ids

        # Add all the features to the preprocessed data
        pre_training_preprocessed.append(
            {
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "token_type_ids": token_type_ids,
                "masked_lm_labels": masked_lm_labels,
                "ent_masked_lm_labels": ent_masked_lm_labels,
                "next_sentence_labels": 0,  # 0 indicates the random walks belong to the text evidence
            }
        )

    # Put the preprocessed data into a dataframe
    pre_training_preprocessed_df = pd.DataFrame(pre_training_preprocessed)

    return pre_training_preprocessed_df


def get_stonkgs_embeddings(
    preprocessed_df: pd.DataFrame,
    pretrained_stonkgs_model_name: Optional[str],
    list_of_indices: Optional[List],
) -> pd.DataFrame:
    """Extract embeddings for a preprocessed_df based on a pretrained_stonkgs_model_name."""
    all_embed_sequences = pd.DataFrame(columns=["embedding"])

    # Initialize the model
    if pretrained_stonkgs_model_name:
        stonkgs_model = STonKGsForPreTraining.from_pretrained(pretrained_stonkgs_model_name)
    else:
        stonkgs_model = STonKGsForPreTraining.from_default_pretrained()

    # Use all indices for extraction if they are unspecified
    if list_of_indices is None:
        list_of_indices = list(range(len(preprocessed_df)))

    for idx in tqdm(list_of_indices):
        data_entry = {
            key: torch.tensor([value]) for key, value in dict(preprocessed_df.iloc[idx]).items()
        }
        stonkgs_hidden_states = stonkgs_model(**data_entry, return_dict=True).pooler_output[0]
        all_embed_sequences = all_embed_sequences.append(
            {"embedding": stonkgs_hidden_states.tolist()},
            ignore_index=True,
        )

    return all_embed_sequences



# -*- coding: utf-8 -*-

"""Reads and prepares triples/statements for INDRA.

Run with:
python -m src.stonkgs.data.indra_extraction
"""

import json
import logging
import os
from typing import Any, Dict, List, Tuple

import networkx as nx
import pandas as pd
import pybel
from pybel.constants import (
    ANNOTATIONS,
    EVIDENCE,
    RELATION,
    CITATION,
    INCREASES,
    DIRECTLY_INCREASES,
    DECREASES,
    DIRECTLY_DECREASES,
    REGULATES,
    BINDS,
    CORRELATION,
    NO_CORRELATION,
    NEGATIVE_CORRELATION,
    POSITIVE_CORRELATION,
    ASSOCIATION,
    PART_OF,
)
from pybel.dsl import (
    CentralDogma,
    ComplexAbundance,
    Abundance,
    CompositeAbundance,
    MicroRna,
    BaseConcept,
    ListAbundance,
    Reaction,
)
from tqdm import tqdm

from stonkgs.constants import (
    INDRA_RAW_JSON,
    MISC_DIR,
    PRETRAINING_DIR,
    SPECIES_DIR,
    CELL_LINE_DIR,
    LOCATION_DIR,
    DISEASE_DIR,
    RELATION_TYPE_DIR,
)

logger = logging.getLogger(__name__)

DIRECT_RELATIONS = {
    DIRECTLY_INCREASES,
    DIRECTLY_DECREASES,
    BINDS,
}

INDIRECT_RELATIONS = {
    REGULATES,
    CORRELATION,
    DECREASES,
    INCREASES,
    NO_CORRELATION,
    NEGATIVE_CORRELATION,
    POSITIVE_CORRELATION,
    ASSOCIATION,
    PART_OF,
}

UP_RELATIONS = {INCREASES, POSITIVE_CORRELATION, DIRECTLY_INCREASES}

DOWN_RELATIONS = {DECREASES, NEGATIVE_CORRELATION, DIRECTLY_DECREASES}


def binarize_triple_direction(
    graph: pybel.BELGraph, triples_per_class: int = 25000
) -> Tuple[Dict[str, Any], List]:
    """Binarize triples depending on the type of direction.

    Extract the fine-tuning data for the interaction type (direct vs. indirect) and polarity (up vs. down) tasks.
    """
    triples = []

    edges_to_removes = []

    counter_dir_inc = 0
    counter_dir_dec = 0
    counter_inc = 0
    counter_dec = 0

    summary = {"context": "(in)direct relations and polarity"}

    # Iterate through the graph and infer a subgraph
    for u, v, k, data in graph.edges(keys=True, data=True):

        if EVIDENCE not in data or not data[EVIDENCE] or data[EVIDENCE] == "No evidence text.":
            # logger.warning(f'not evidence found in {data}')
            continue

        # Both nodes in the triple are required to be a protein/gene (complexes and other stuff are skipped)
        if not isinstance(u, CentralDogma) and not isinstance(v, CentralDogma):
            continue

        if data[RELATION] in UP_RELATIONS:
            polarity_label = "up"
        elif data[RELATION] in DOWN_RELATIONS:
            polarity_label = "down"
        else:
            continue

        if data[RELATION] in {INCREASES, DECREASES}:
            interaction_label = "indirect_interaction"
        elif data[RELATION] in {DIRECTLY_INCREASES, DIRECTLY_DECREASES}:
            interaction_label = "direct_interaction"
        else:
            continue

        """Check if limit has been reached"""
        if data[RELATION] == DIRECTLY_DECREASES and counter_dir_dec >= triples_per_class:
            continue
        elif data[RELATION] == INCREASES and counter_inc >= triples_per_class:
            continue
        elif data[RELATION] == DECREASES and counter_dec >= triples_per_class:
            continue
        elif data[RELATION] == DIRECTLY_INCREASES and counter_dir_inc >= triples_per_class:
            continue

        # Add particular triple to the fine tuning set
        if data[RELATION] == INCREASES:
            counter_inc += 1
        elif data[RELATION] == DIRECTLY_INCREASES:
            counter_dir_inc += 1
        elif data[RELATION] == DIRECTLY_DECREASES:
            counter_dir_dec += 1
        elif data[RELATION] == DECREASES:
            counter_dec += 1
        else:
            continue

        triples.append(
            {
                "source": u,
                "relation": data[RELATION],
                "target": v,
                "evidence": data[EVIDENCE],
                "pmid": data[CITATION],
                "polarity": polarity_label,
                "interaction": interaction_label,
            }
        )

        edges_to_removes.append((u, v, k))

    df = pd.DataFrame(triples)

    logger.info(f"Number of binarized triples for fine-tuning: {df.shape[0]}")

    summary["number_of_triples"] = df.shape[0]
    summary["number_of_labels"] = "4 or 2 depending on the task"
    summary["labels"] = "NA"

    df.to_csv(os.path.join(RELATION_TYPE_DIR, f"relation_type.tsv"), sep="\t", index=False)

    return summary, edges_to_removes


def create_polarity_annotations(graph: pybel.BELGraph) -> Dict[str, Any]:
    """Group triples depending on the type of polarity."""
    triples = []

    summary = {"context": "polarity"}

    # Iterate through the graph and infer a subgraph
    for u, v, data in graph.edges(data=True):

        if EVIDENCE not in data or not data[EVIDENCE] or data[EVIDENCE] == "No evidence text.":
            logger.warning(f"not evidence found in {data}")
            continue

        # todo: check this we will focus only on molecular interactions
        if not any(
            isinstance(u, class_to_check)
            for class_to_check in (
                CentralDogma,
                ComplexAbundance,
                Abundance,
                CompositeAbundance,
                MicroRna,
            )
        ):
            continue

        if not any(
            isinstance(v, class_to_check)
            for class_to_check in (
                CentralDogma,
                ComplexAbundance,
                Abundance,
                CompositeAbundance,
                MicroRna,
            )
        ):
            continue

        class_label = "indirect" if data[RELATION] in INDIRECT_RELATIONS else "direct"

        triples.append(
            {
                "source": u,
                "relation": data[RELATION],
                "target": v,
                "evidence": data[EVIDENCE],
                "pmid": data[CITATION],
                "class": class_label,
            }
        )

    df = pd.DataFrame(triples)

    summary["number_of_triples"] = df.shape[0]
    summary["number_of_labels"] = df["class"].unique().size
    summary["labels"] = df["class"].value_counts().to_dict()

    df.to_csv(os.path.join(RELATION_TYPE_DIR, f"relation_type.tsv"), sep="\t", index=False)

    return summary


def create_context_type_specific_subgraph(
    graph: pybel.BELGraph,
    context_annotations: List[str],
) -> Tuple[List, pybel.BELGraph]:
    """Create a subgraph based on context annotations and also return edges that should be removed later on."""
    subgraph = graph.child()
    subgraph.name = f"INDRA graph contextualized for {context_annotations}"

    edges_to_remove = []

    # Iterate through the graph and infer a subgraph with edges that contain the annotation of interest
    for u, v, k, data in graph.edges(data=True, keys=True):
        if ANNOTATIONS in data and any(
            annotation in data[ANNOTATIONS] for annotation in context_annotations
        ):
            subgraph.add_edge(u, v, k, **data)
            # Triples to be removed
            edges_to_remove.append((u, v, k))

    # number_of_edges_before = graph.number_of_edges()
    # graph.remove_edges_from(edges_to_remove)
    # number_of_edges_after_removing_annotations = graph.number_of_edges()

    # logger.info(
    #     f'Original graph was reduced from {number_of_edges_before} to {number_of_edges_after_removing_annotations}
    #     edges'
    # )

    string = f'Number of nodes/edges in the inferred subgraph "{context_annotations}": \
    {subgraph.number_of_nodes()} {subgraph.number_of_edges()}'

    logger.info(string)

    return edges_to_remove, subgraph


def dump_edgelist(
    graph: pybel.BELGraph,
    annotations: List[str],
    name: str,
    output_dir: str,
) -> Dict[str, Any]:
    """Dump tsv file for ml purposes."""
    triples = []

    summary = {
        "context": name,
    }

    # Iterate through the graph and infer a subgraph with edges that contain the annotation of interest
    for u, v, data in graph.edges(data=True):

        # If the data entry has no text evidence or the following filler text, don't add it
        if not data[EVIDENCE] or data[EVIDENCE] == "No evidence text.":
            continue

        # Multiple annotations
        for annotation in data[ANNOTATIONS]:

            if annotation not in annotations:
                continue

            # Skip multiple classes in the triple for the same annotation
            if len(data[ANNOTATIONS][annotation]) > 1:
                logger.warning(f"triple has more than one label -> {data[ANNOTATIONS][annotation]}")
                continue

            for label_annotation in data[ANNOTATIONS][annotation]:
                triples.append(
                    {
                        "source": u,
                        "relation": data[RELATION],
                        "target": v,
                        "evidence": data[EVIDENCE],
                        "pmid": data[CITATION],
                        "class": label_annotation,
                    },
                )

    if not triples:
        return {
            "context": name,
            "number_of_triples": "0",
            "number_of_labels": "0",
            "labels": "0",
        }

    df = pd.DataFrame(triples)

    # Leave out any filtering for now, keep all the classes
    """Removing labels that appear less than X% from the total"""
    """    label_counts = df['class'].value_counts().to_dict()
    # 0.05%
    percentage = 0.05
    cutoff = int(df.shape[0] * percentage)

    labels_to_remove = {
        label: count
        for label, count in label_counts.items()
        if count < cutoff
    }

    logger.warning(f'labels for {name} removed due to low occurrence {labels_to_remove}')

    logger.info(f'raw triples {df.shape[0]}')
    df = df[~df['class'].isin(list(labels_to_remove.keys()))]"""

    logger.info(f"triples (no filtering) for {name}: {df.shape[0]}")

    final_labels = df["class"].unique()
    logger.info(f" final number of classes {final_labels.size}")

    summary["number_of_triples"] = df.shape[0]
    summary["number_of_labels"] = final_labels.size
    summary["labels"] = df["class"].value_counts().to_dict()

    df.to_csv(os.path.join(output_dir, f"{name}.tsv"), sep="\t", index=False)

    return summary


def munge_evidence_text(text: str) -> str:
    """Clean evidence."""
    # Deal with the xrefs in text
    if "XREF_BIBR" in text:
        text = text.replace("XREF_BIBR, ", "")
        text = text.replace("XREF_BIBR,", "")
        text = text.replace("XREF_BIBR", "")
        text = text.replace("[", "")
        text = text.replace("]", "")

    return text


def read_indra_triples(
    path: str = INDRA_RAW_JSON,
    batch_processing: bool = False,
    batch_size: int = 10000000,
):
    """Parse indra statements in JSON and returns context specific graphs."""
    #: Read file INDRA KG
    errors = []
    lines = []

    with open(path) as file:
        for line_number, line in tqdm(
            enumerate(file),
            desc="parsing file",
            total=35150093,  # TODO: hard coded
        ):
            try:
                line_dict = json.loads(line)
            except:
                errors.append(line_number)

            lines.append(line_dict)

    logger.info(f"{len(errors)} statements with errors from {len(lines)} statements")

    if batch_processing:
        # round down the number of chunks
        chunks = len(lines) // batch_size

        # create a list for the partial KGs that should be merged in the end
        partial_indra_kgs = []
        for i in tqdm(range(chunks), total=chunks, desc="processing partial KGs"):
            # process the lines chunk wise
            partial_indra_kgs.append(
                pybel.io.indra.from_indra_statements_json(
                    lines[i * batch_size : (i + 1) * batch_size]
                )
            )
        # process last chunk differently
        partial_indra_kgs.append(
            pybel.io.indra.from_indra_statements_json(lines[(i + 1) * batch_size :])
        )

        logger.info(f"Finished processing {chunks + 1} many chunks")

        indra_kg = pybel.union(partial_indra_kgs)

        del partial_indra_kgs

    else:
        indra_kg = pybel.io.indra.from_indra_statements_json(lines)

    # Remove non grounded nodes
    non_grounded_nodes = {
        node
        for node in indra_kg.nodes()
        if isinstance(node, BaseConcept) and node.curie.startswith("TEXT:")
    }

    # Remove non grounded nodes from complex nodes
    for node in indra_kg.nodes():
        # Process Complex/Composites
        if isinstance(node, ListAbundance):
            for member in node.members:
                if isinstance(member, BaseConcept) and member.curie.startswith("TEXT:"):
                    non_grounded_nodes.add(node)

        # Process Reactions
        if isinstance(node, Reaction):
            for member in node.reactants:
                if isinstance(member, BaseConcept) and member.curie.startswith("TEXT:"):
                    non_grounded_nodes.add(node)

            for member in node.products:
                if isinstance(member, BaseConcept) and member.curie.startswith("TEXT:"):
                    non_grounded_nodes.add(node)

    logger.warning(f"removing {len(non_grounded_nodes)} non grounded nodes")

    indra_kg.remove_nodes_from(non_grounded_nodes)

    # Remove any nodes that are not in the largest component
    connected_components = [
        component
        for component in sorted(
            nx.connected_components(indra_kg.to_undirected()), key=len, reverse=True
        )
    ]
    logger.warning(f"Components of the KG: {len(connected_components)}")
    connected_components_size = [len(c) for c in connected_components]
    logger.warning(
        f"Largest component {connected_components_size[0]}, first 50... -> {connected_components_size[1: 50]}"
    )
    nodes_not_in_largest_component = [
        node for subgraph in connected_components[1:] for node in subgraph
    ]
    indra_kg.remove_nodes_from(nodes_not_in_largest_component)

    logger.warning(
        f"{len(nodes_not_in_largest_component)} nodes were removed because they are not in the largest "
        f"component"
    )

    #: Summarize content of the KG
    logger.info(f"{indra_kg.number_of_edges()} edges from {len(lines)} statements")
    logger.info(indra_kg.summarize())

    # Dump the most important characteristics of the summary to a file
    summary = {
        "node_summary": indra_kg.count.namespaces(),
        "relation_summary": indra_kg.count.relations(),
        "functions_summary": indra_kg.count.functions(),
        "annotations_summary": indra_kg.count.annotations(),
    }
    summary_list = [{"name": key, "value": value} for key, value in summary.items()]

    with open(os.path.join(MISC_DIR, "indra_kg_overview_summary.json"), "w") as f:
        json.dump(summary_list, f, ensure_ascii=False)

    # Print all the annotations in the graph
    all_annotations = set()

    for _, _, data in indra_kg.edges(data=True):
        if ANNOTATIONS in data:
            for key in data[ANNOTATIONS]:
                all_annotations.add(key)

    # Summarize all annotations
    logger.info(f"all annotations -> {all_annotations}")

    """
    Split the KG into two big chunks:

    1. Pre-training dataset -> Any triple + text evidence that is not in any of the 4+2 annotation types above
    The rationale is that we want to have a common pre-trained model that can be used as a basis for all the fine tuning
    classification tasks.

    2. Fine-tuning datasets (benchmark for our models). 6 different train-validation-test splits for each of the 
    annotation types (each of them will contain multiple classes).

    We would like to note that the pre-training dataset is significantly larger than the fine-tuning dataset.
    Naturally, STonKGs requires a pre-training similar to the other two baselines models (i.e., KG-based has been
    trained based on node2vec on the INDRA KG and BioBERT was trained on PubMed).
    """
    species_edges, species_subgraph = create_context_type_specific_subgraph(indra_kg, ["species"])
    disease_edges, disease_subgraph = create_context_type_specific_subgraph(indra_kg, ["disease"])
    cell_line_edges, cell_line_subgraph = create_context_type_specific_subgraph(
        indra_kg, ["cell_line"]
    )
    location_edges, location_subgraph = create_context_type_specific_subgraph(
        indra_kg, ["location"]
    )

    #: Dump the 4+2 annotation type specific subgraphs (triples)
    species_summary = dump_edgelist(
        graph=species_subgraph,
        annotations=["species"],
        name="species",
        output_dir=SPECIES_DIR,
    )
    disease_summary = dump_edgelist(
        graph=disease_subgraph,
        annotations=["disease"],
        name="disease",
        output_dir=DISEASE_DIR,
    )
    cell_line_summary = dump_edgelist(
        graph=cell_line_subgraph,
        annotations=["cell_line"],
        name="cell_line",
        output_dir=CELL_LINE_DIR,
    )
    location_summary = dump_edgelist(
        graph=location_subgraph,
        annotations=["location"],
        name="location",
        output_dir=LOCATION_DIR,
    )

    polarity_summary, polarity_edges = binarize_triple_direction(indra_kg)

    summary_df = pd.DataFrame(
        [
            species_summary,
            disease_summary,
            cell_line_summary,
            location_summary,
            polarity_summary,  # This is actually two tasks (polarity and direction)
        ]
    )
    summary_df.to_csv(os.path.join(MISC_DIR, "summary.tsv"), sep="\t", index=False)

    # Remove all the fine-tuning edges from the pre-training data
    for edges in [
        species_edges,
        disease_edges,
        cell_line_edges,
        location_edges,
        polarity_edges,
    ]:
        indra_kg.remove_edges_from(edges)

    """Dump pre-training dataset."""
    triples = []

    # Iterate through the graph and infer a subgraph with edges that contain the annotation of interest
    for u, v, data in tqdm(indra_kg.edges(data=True), desc="Building final pre-training dataframe"):
        # Skip relations without evidences
        if EVIDENCE not in data or data[EVIDENCE] == "No evidence text.":
            continue

        triples.append(
            {
                "source": u,
                "relation": data[RELATION],
                "target": v,
                "evidence": munge_evidence_text(data[EVIDENCE]),
                "pmid": data[CITATION],
                "belief_score": data[ANNOTATIONS].get("belief", ""),
            }
        )

    pretraining_triples = pd.DataFrame(triples)
    del triples
    pretraining_triples.to_csv(
        os.path.join(PRETRAINING_DIR, "pretraining_triples.tsv"), sep="\t", index=False
    )


if __name__ == "__main__":
    read_indra_triples()



# -*- coding: utf-8 -*-

"""Runs the STonKGs model on the fine-tuning classification task, assuming the model embeddings are pre-trained.

Run with:
python -m src.stonkgs.models.stonkgs_finetuning
"""

import logging
import os
from collections import Counter
from typing import Dict, List, Optional

import click  # type: ignore
import mlflow
import numpy as np
import pandas as pd
import torch
from sklearn.metrics import f1_score
from sklearn.model_selection import KFold, StratifiedShuffleSplit
from tqdm import tqdm
from transformers.modeling_outputs import SequenceClassifierOutput
from transformers.models.bert import BertModel, BertTokenizer, BertTokenizerFast
from transformers.trainer import Trainer, TrainingArguments

from stonkgs.constants import (
    CELL_LINE_DIR,
    CORRECT_DIR,
    DEEPSPEED_CONFIG_PATH,
    DISEASE_DIR,
    EMBEDDINGS_PATH,
    LOCATION_DIR,
    MLFLOW_FINETUNING_TRACKING_URI,
    NDD_DIR,
    NLP_MODEL_TYPE,
    PRETRAINED_STONKGS_DUMMY_PATH,
    RANDOM_WALKS_PATH,
    RELATION_TYPE_DIR,
    SPECIES_DIR,
    STONKGS_OUTPUT_DIR,
    VOCAB_FILE,
)
from stonkgs.models.kg_baseline_model import prepare_df
from stonkgs.models.stonkgs_model import STonKGsForPreTraining

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
# Disable alembic info
logging.getLogger("alembic").setLevel(logging.WARNING)


def get_train_test_splits(
    train_data: pd.DataFrame,
    type_column_name: str = "labels",
    random_seed: int = 42,
    n_splits: int = 5,
    max_dataset_size: int = 100000,
) -> List:
    """Return train/test indices for n_splits many splits based on the fine-tuning dataset that is passed."""
    # Leave out the label in the dataset
    data = train_data.drop(type_column_name, axis=1)
    labels = train_data[type_column_name]

    # Cut the dataset down to max_dataset_size (deterministically!) using StratifiedShuffleSplit if needed:
    # (this is not an actual train/test split, this is just for getting a dataset of size max_dataset_size in a
    # stratified and deterministic manner)
    if len(data) > max_dataset_size:
        splitter = StratifiedShuffleSplit(
            n_splits=1,
            train_size=max_dataset_size,
            random_state=random_seed,
        )
        for train_index, _ in splitter.split(data, labels):
            data = data.iloc[train_index, :].reset_index(drop=True)
            labels = labels.iloc[train_index].reset_index(drop=True)

    # Generate the actual train/test splits here:
    # Implement non-stratified train/test splits
    skf = KFold(n_splits=n_splits, random_state=random_seed, shuffle=True)

    return [
        {"train_idx": train_idx, "test_idx": test_idx}
        for train_idx, test_idx in skf.split(data, labels)
    ]


def preprocess_fine_tuning_data(
    train_data_path: str,
    class_column_name: str = "class",
    embedding_name_to_vector_path: str = EMBEDDINGS_PATH,
    embedding_name_to_random_walk_path: str = RANDOM_WALKS_PATH,
    nlp_model_type: str = NLP_MODEL_TYPE,
    sep_id: int = 102,
    unk_id: int = 100,
) -> pd.DataFrame:
    """Generate input_ids, attention_mask, token_type_ids etc. based on the source, target, evidence columns."""
    # Load the KG embedding dict to convert the names to numeric indices
    kg_embed_dict = prepare_df(embedding_name_to_vector_path)
    kg_name_to_idx = {key: i for i, key in enumerate(kg_embed_dict.keys())}
    # Load the random walks for each node
    random_walk_dict = prepare_df(embedding_name_to_random_walk_path)
    # Convert random walk sequences to list of numeric indices
    random_walk_idx_dict = {
        k: [kg_name_to_idx[node] for node in v] for k, v in random_walk_dict.items()
    }

    # Load the raw fine-tuning dataset with source, target and evidence
    unprocessed_df = pd.read_csv(
        train_data_path, sep="\t", usecols=["source", "target", "evidence", class_column_name]
    )

    # TODO: leave it out later on?
    # Filter out any triples that contain a node that is not in the embeddings_dict
    original_length = len(unprocessed_df)
    unprocessed_df = unprocessed_df[
        unprocessed_df["source"].isin(kg_embed_dict.keys())
        & unprocessed_df["target"].isin(kg_embed_dict.keys())
    ].reset_index(drop=True)
    new_length = len(unprocessed_df)
    logger.info(
        f"{original_length - new_length} out of {original_length} triples are left out because they contain "
        f"nodes which are not present in the pre-training data"
    )

    # Check how many nodes in the fine-tuning dataset are not covered by the learned KG embeddings
    number_of_pre_training_nodes = len(
        set(unprocessed_df["source"]).union(set(unprocessed_df["target"]))
    )
    if number_of_pre_training_nodes > len(kg_embed_dict):
        logger.warning(
            f"{number_of_pre_training_nodes - len(kg_embed_dict)} out of {number_of_pre_training_nodes}"
            f"nodes are not covered by the embeddings learned in the pretraining dataset"
        )

    # Get the length of the text or entity embedding sequences (2 random walks + 2 = entity embedding sequence length)
    random_walk_length = len(next(iter(random_walk_idx_dict.values())))
    half_length = random_walk_length * 2 + 2

    # Initialize a FAST tokenizer if it's the default one (BioBERT)
    if nlp_model_type == NLP_MODEL_TYPE:
        # Initialize the fast tokenizer for getting the text token ids
        tokenizer = BertTokenizerFast(vocab_file=VOCAB_FILE)
    else:
        # Initialize a slow tokenizer used for getting the text token ids
        tokenizer = BertTokenizer.from_pretrained(nlp_model_type)

    # Initialize the preprocessed data
    fine_tuning_preprocessed = []

    # Log progress with a progress bar
    for _, row in tqdm(
        unprocessed_df.iterrows(),
        total=unprocessed_df.shape[0],
        desc="Preprocessing the fine-tuning dataset",
    ):
        # 1. "Token type IDs": 0 for text tokens, 1 for entity tokens
        token_type_ids = [0] * half_length + [1] * half_length

        # 2. Tokenization for getting the input ids and attention masks for the text
        # Use encode_plus to also get the attention mask ("padding" mask)
        encoded_text = tokenizer.encode_plus(
            row["evidence"],
            padding="max_length",
            truncation=True,
            max_length=half_length,
        )
        text_token_ids = encoded_text["input_ids"]
        text_attention_mask = encoded_text["attention_mask"]

        # 3. Get the random walks sequence and the node indices, add the SEP (usually with id=102) in between
        # Use a sequence of UNK tokens if the node is not contained in the dictionary of the nodes from pre-training
        random_w_source = (
            random_walk_idx_dict[row["source"]]
            if row["source"] in random_walk_idx_dict.keys()
            else [unk_id] * random_walk_length
        )
        random_w_target = (
            random_walk_idx_dict[row["target"]]
            if row["target"] in random_walk_idx_dict.keys()
            else [unk_id] * random_walk_length
        )
        random_w_ids = random_w_source + [sep_id] + random_w_target + [sep_id]

        # 4. Total attention mask (attention mask is all 1 for the entity sequence)
        attention_mask = text_attention_mask + [1] * half_length

        # 5. Total input_ids = half text ids + half entity ids
        input_ids = text_token_ids + random_w_ids

        # Add all the features to the preprocessed data
        fine_tuning_preprocessed.append(
            {
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "token_type_ids": token_type_ids,  # Remove the MLM, ELM and NSP labels since it's not needed anymore
                "labels": row[
                    class_column_name
                ],  # Add the annotation/relation label for fine-tuning instead
            }
        )

    # Put the preprocessed data into a dataframe
    fine_tuning_preprocessed_df = pd.DataFrame(fine_tuning_preprocessed)

    return fine_tuning_preprocessed_df


class INDRAEntityEvidenceDataset(torch.utils.data.Dataset):
    """Custom Dataset class for INDRA data containing the combination of text and KG triple data."""

    def __init__(
        self,
        encodings,
        labels,
    ):
        """Initialize INDRA Dataset based on the combined input sequence consisting of text and triple data."""
        self.encodings = encodings
        # Assumes that the labels are numerically encoded
        self.labels = labels

    def __getitem__(self, idx):
        """Return data entries for given indices."""
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        """Return the length of the dataset."""
        return len(self.labels)


class STonKGsForSequenceClassification(STonKGsForPreTraining):
    """Create the fine-tuning part of the STonKGs model based the pre-trained STonKGs model.

    Note that this class inherits from STonKGsForPreTraining rather than PreTrainedModel, thereby it's deviating from
    the typical huggingface inheritance logic of the fine-tuning classes.
    """

    def __init__(self, config):
        """Initialize the STonKGs sequence classification model based on the pre-trained STonKGs model architecture."""
        super().__init__(config)
        self.num_labels = config.num_labels
        self.config = config

        # Load the pretrained STonKGs Transformer here
        self.bert = BertModel(config)
        self.dropout = torch.nn.Dropout(config.hidden_dropout_prob)
        self.classifier = torch.nn.Linear(config.hidden_size, config.num_labels)

        # Initialize all the pre-trained as well as new weights (i.e. classifier weights) here
        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        """Perform one forward pass for a given sequence of text_input_ids + ent_input_ids."""
        # Use the LM backbone to get the pre-trained token embeddings
        # batch x half_length x hidden_size
        # The first element of the returned tuple from the LM backbone forward() pass is the sequence of hidden states
        token_embeddings = self.lm_backbone(input_ids[:, : self.cls.predictions.half_length])[0]

        # Use the KG backbone to obtain the pre-trained entity embeddings
        # batch x half_length x hidden_size
        ent_embeddings = torch.stack(
            [
                # for each numeric index in the random walks sequence: get the embedding vector from the KG backbone
                torch.stack([self.kg_backbone[i.item()] for i in j])
                # for each example in the batch: get the random walks sequence
                for j in input_ids[:, self.cls.predictions.half_length :]
            ],
        )

        # Concatenate token and entity embeddings obtained from the LM and KG backbones and cast to float
        # batch x seq_len x hidden_size
        inputs_embeds = (
            torch.cat(
                [token_embeddings, ent_embeddings.to(token_embeddings.device)],
                dim=1,
            )
            .type(torch.FloatTensor)
            .to(self.device)
        )

        # Get the hidden states from the pretrained STonKGs Transformer layers
        outputs = self.bert(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            head_mask=head_mask,
            return_dict=None,
        )
        # Only use the pooled output (of the [CLS] token)
        pooled_output = outputs[1]

        # Apply dropout and the linear layer
        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        loss = None
        if labels is not None:
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (
                    labels.dtype == torch.long or labels.dtype == torch.int
                ):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = torch.nn.MSELoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = torch.nn.CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = torch.nn.BCEWithLogitsLoss()
                loss = loss_fct(logits, labels)
        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


def run_sequence_classification_cv(
    train_data_path: str,
    model_path: str = PRETRAINED_STONKGS_DUMMY_PATH,
    output_dir: Optional[str] = STONKGS_OUTPUT_DIR,
    logging_uri_mlflow: Optional[str] = MLFLOW_FINETUNING_TRACKING_URI,
    label_column_name: str = "labels",
    class_column_name: str = "class",
    epochs: Optional[int] = 10,
    log_steps: int = 500,
    lr: float = 5e-5,
    batch_size: int = 8,
    gradient_accumulation: int = 1,
    task_name: str = "",
    deepspeed: bool = True,
    max_dataset_size: int = 100000,
) -> Dict:
    """Run cross-validation for the sequence classification task(s) using STonKGs."""
    # Get data splits
    fine_tuning_df = preprocess_fine_tuning_data(
        train_data_path=train_data_path,
        class_column_name=class_column_name,
    )

    train_test_splits = get_train_test_splits(
        fine_tuning_df,
        max_dataset_size=max_dataset_size,
    )

    # Get text evidences and labels
    fine_tuning_data, labels_str = (
        fine_tuning_df.drop(columns=label_column_name),
        fine_tuning_df[label_column_name],
    )
    # Numerically encode labels
    unique_tags = set(label for label in labels_str)
    tag2id = {label: number for number, label in enumerate(unique_tags)}
    id2tag = {value: key for key, value in tag2id.items()}
    labels = pd.Series([int(tag2id[label]) for label in labels_str])

    # Initialize the f1-score
    f1_scores = []

    # End previous run
    mlflow.end_run()
    # Initialize mlflow run, set tracking URI to use the same experiment for all runs,
    # so that one can compare them
    mlflow.set_tracking_uri(logging_uri_mlflow)
    mlflow.set_experiment("STonKGs Fine-Tuning")

    # Initialize a dataframe for all the predicted labels
    result_df = pd.DataFrame()

    for idx, indices in enumerate(train_test_splits):
        model = STonKGsForSequenceClassification.from_pretrained(
            pretrained_model_name_or_path=model_path,
            num_labels=len(unique_tags),
        )

        # Based on the preprocessed fine-tuning dataframe: Convert the data into the desired dictionary format
        # for the INDRAEntityEvidenceDataset
        train_data = (
            fine_tuning_data.iloc[indices["train_idx"]]
            .reset_index(drop=True)
            .to_dict(orient="list")
        )
        test_data = (
            fine_tuning_data.iloc[indices["test_idx"]].reset_index(drop=True).to_dict(orient="list")
        )
        train_labels = labels[indices["train_idx"]].tolist()
        test_labels = labels[indices["test_idx"]].tolist()
        train_dataset = INDRAEntityEvidenceDataset(encodings=train_data, labels=train_labels)
        test_dataset = INDRAEntityEvidenceDataset(encodings=test_data, labels=test_labels)

        # Note that due to the randomization in the batches, the training/evaluation is slightly
        # different every time
        # TrainingArgument uses a default batch size of 8
        training_args = TrainingArguments(
            output_dir=output_dir,
            num_train_epochs=epochs,  # total number of training epochs
            logging_steps=log_steps,  # reduce the number of logging steps to avoid collisions when writing to the
            # shared database
            # Use deepspeed with a specified config file for speedup
            deepspeed=DEEPSPEED_CONFIG_PATH if deepspeed else None,
            learning_rate=lr,
            report_to=["mlflow"],  # log via mlflow
            do_train=True,
            do_predict=True,
            per_device_train_batch_size=batch_size,
            gradient_accumulation_steps=gradient_accumulation,
        )

        # Initialize Trainer based on the training dataset
        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
        )
        # Train
        trainer.train()

        # Log some details about the datasets used in training and testing
        mlflow.log_param("label dict", str(tag2id))
        mlflow.log_param("training dataset size", str(len(train_labels)))
        mlflow.log_param("training class dist", str(Counter(train_labels)))
        mlflow.log_param("test dataset size", str(len(test_labels)))
        mlflow.log_param("test class dist", str(Counter(test_labels)))

        # Make predictions for the test dataset
        predictions = trainer.predict(test_dataset=test_dataset).predictions
        predicted_labels = np.argmax(predictions, axis=1)
        logger.info(f"Predicted labels: {predicted_labels}")

        # Save the predicted + true labels
        partial_result_df = pd.DataFrame(
            {
                "split": idx,
                "index": indices["test_idx"].tolist(),
                "predicted_label": predicted_labels.tolist(),
                "true_label": test_labels,
                "evidence": fine_tuning_data.iloc[indices["test_idx"]]["input_ids"].tolist(),
            },
        )
        result_df = result_df.append(
            partial_result_df,
            ignore_index=True,
        )

        # Use weighted average
        f1_sc = f1_score(test_labels, predicted_labels, average="weighted")
        f1_scores.append(f1_sc)

        # Log the final f1_score
        mlflow.log_metric("f1_score_weighted", f1_sc)

    # Log mean and std f1-scores from the cross validation procedure (average and std across all splits) to the
    # standard logger
    logger.info(f"Mean f1-score: {np.mean(f1_scores)}")
    logger.info(f"Std f1-score: {np.std(f1_scores)}")

    # Map the labels in the result df back to their original names
    result_df = result_df.replace({"predicted_label": id2tag, "true_label": id2tag})
    # Save the result_df
    result_df.to_csv(
        os.path.join(STONKGS_OUTPUT_DIR, "predicted_labels_stonkgs_" + task_name + "df.tsv"),
        index=False,
        sep="\t",
    )

    # Save the last model
    trainer.save_model(output_dir=os.path.join(STONKGS_OUTPUT_DIR, task_name))

    # End the previous run
    mlflow.end_run()

    # Log the mean and std f1 score from the cross validation procedure to mlflow
    with mlflow.start_run():
        # Log the task name as well
        mlflow.log_param("task name", task_name)
        mlflow.log_metric("f1_score_mean", np.mean(f1_scores))
        mlflow.log_metric("f1_score_std", np.std(f1_scores))

    return {"f1_score_mean": np.mean(f1_scores), "f1_score_std": np.std(f1_scores)}


@click.command()
@click.option("-e", "--epochs", default=5, help="Number of epochs", type=int)
@click.option("--lr", default=5e-5, help="Learning rate", type=float)
@click.option(
    "--logging_dir",
    default=MLFLOW_FINETUNING_TRACKING_URI,
    help="Mlflow logging/tracking URI",
    type=str,
)
@click.option("--log_steps", default=500, help="Number of steps between each log", type=int)
@click.option(
    "--model_path",
    default=PRETRAINED_STONKGS_DUMMY_PATH,
    help="Path of the pretrained model",
    type=str,
)
@click.option("--output_dir", default=STONKGS_OUTPUT_DIR, help="Output directory", type=str)
@click.option("--batch_size", default=8, help="Batch size used in fine-tuning", type=int)
@click.option(
    "--gradient_accumulation_steps", default=1, help="Gradient accumulation steps", type=int
)
@click.option("--deepspeed", default=True, help="Whether to use deepspeed or not", type=bool)
@click.option(
    "--max_dataset_size",
    default=100000,
    help="Maximum dataset size of the fine-tuning datasets",
    type=int,
)
@click.option("--local_rank", default=-1, help="THIS PARAMETER IS IGNORED", type=int)
def run_all_fine_tuning_tasks(
    epochs: int = 5,
    log_steps: int = 500,
    lr: float = 5e-5,
    model_path: str = PRETRAINED_STONKGS_DUMMY_PATH,
    output_dir: str = STONKGS_OUTPUT_DIR,
    logging_dir: Optional[str] = MLFLOW_FINETUNING_TRACKING_URI,
    batch_size: int = 8,
    gradient_accumulation_steps: int = 1,
    deepspeed: bool = True,
    max_dataset_size: int = 100000,  # effectively removes the max dataset size restriction
    local_rank: int = -1,
):
    """Run all fine-tuning tasks at once."""
    # Specify all directories and file names
    directories = [
        CELL_LINE_DIR,
        CORRECT_DIR,
        CORRECT_DIR,
        DISEASE_DIR,
        LOCATION_DIR,
        SPECIES_DIR,
        RELATION_TYPE_DIR,
        RELATION_TYPE_DIR,
        NDD_DIR,
    ]
    file_names = [
        "cell_line_no_duplicates.tsv",
        "correct_incorrect_binary_no_duplicates.tsv",
        "correct_incorrect_multiclass_no_duplicates.tsv",
        "disease_no_duplicates.tsv",
        "location_no_duplicates.tsv",
        "species_no_duplicates.tsv",
        "relation_type_no_duplicates.tsv",
        "relation_type_no_duplicates.tsv",
        "ndd_no_duplicates.tsv",
    ]
    task_names = [
        "cell_line",
        "correct_binary",
        "correct_multiclass",
        "disease",
        "location",
        "species",
        "interaction",
        "polarity",
        "ndd",
    ]
    # Specify the column names of the target variable
    column_names = ["class"] * 6 + ["interaction"] + ["polarity"] + ["class"]

    for directory, file, column_name, task_name in zip(
        directories,
        file_names,
        column_names,
        task_names,
    ):
        # TODO comment out later
        if task_name == "ndd":
            # Run the 8 fine-tuning tasks
            run_sequence_classification_cv(
                train_data_path=os.path.join(directory, file),
                model_path=model_path,
                output_dir=output_dir,
                logging_uri_mlflow=logging_dir,
                epochs=epochs,
                log_steps=log_steps,
                lr=lr,
                batch_size=batch_size,
                gradient_accumulation=gradient_accumulation_steps,
                class_column_name=column_name,
                task_name=task_name,
                deepspeed=deepspeed,
                max_dataset_size=max_dataset_size,
            )
            logger.info(f"Finished the {task_name} task")


if __name__ == "__main__":
    # Set the huggingface environment variable for tokenizer parallelism to false
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    # Run all CV fine-tuning tasks
    run_all_fine_tuning_tasks()



# -*- coding: utf-8 -*-

"""STonKGs."""

from .models.stonkgs_finetuning import STonKGsForSequenceClassification
from .models.stonkgs_for_embeddings import get_stonkgs_embeddings, preprocess_df_for_embeddings
from .models.stonkgs_model import STonKGsForPreTraining

__all__ = [
    "get_stonkgs_embeddings",
    "preprocess_df_for_embeddings",
    "STonKGsForPreTraining",
    "STonKGsForSequenceClassification",
]



# -*- coding: utf-8 -*-

"""Node2vec model.

Run with: ``python -m src.stonkgs.models.node2vec``
"""

import logging
import os
import pickle
import random
from typing import Optional, Union

import csrgraph as cg
import networkx as nx
import numpy as np
import optuna
import pandas as pd
from nodevectors import Node2Vec
from optuna.integration import MLflowCallback
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from stellargraph.data import EdgeSplitter
from tqdm import tqdm

from stonkgs.constants import KG_HPO_DIR, MLFLOW_TRACKING_URI, MODELS_DIR, PRETRAINING_PATH

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def run_link_prediction(
    kg: Union[nx.DiGraph, cg.csrgraph],
    model: Node2Vec,
) -> float:
    """Link prediction task for a given KG and node2vec model."""
    # Preprocessing step: Generate positive and negative triples
    # 1. Method employed -> https://stellargraph.readthedocs.io/en/latest/_modules/stellargraph/data/edge_splitter.html
    # 2. Alternative from PyKEEN -> https://pykeen.readthedocs.io/en/latest/reference/negative_sampling.html
    # 3. Alternative from xswap -> https://github.com/hetio/xswap

    if isinstance(kg, cg.csrgraph):
        # The sparse matrix needs to be converted first
        kg_triples = kg.mat  # noqa: F841
        raise NotImplementedError("negative sampling has not yet been implemented")
    else:
        # follow https://stellargraph.readthedocs.io/en/stable/demos/link-prediction/node2vec-link-prediction.html
        _, indra_kg_triples, edge_labels = EdgeSplitter(kg).train_test_split()

    # 1. generate embeddings
    # 2. use dot product between two entities from negative triple
    # iterate row-wise through the negative samples and return the hadamard product of the embeddings of the
    # two entities from each triple
    triple_embeds = np.asarray(
        [model.predict(source) * model.predict(target) for source, target in indra_kg_triples]
    )

    # 3. use a LogReg classifier for final classification on test
    # first generate a train test split for the classifier
    embeds_train, embeds_test, labels_train, labels_test = train_test_split(
        triple_embeds, edge_labels, stratify=edge_labels
    )

    link_prediction_model = LogisticRegression()
    link_prediction_model.fit(embeds_train, labels_train)
    predictions_test = link_prediction_model.predict(embeds_test)

    # 4. return sklearn roc curve score for this classifier
    return roc_auc_score(labels_test, predictions_test)


# TODO: add parameters/click later on
# @click.group()
def run_node2vec_hpo(
    positive_graph_path: Optional[str] = PRETRAINING_PATH,
    sep: Optional[str] = "\t",
    delete_database: Optional[bool] = True,
    mlflow_tracking_uri: Optional[str] = MLFLOW_TRACKING_URI,
    n_optimization_trials: Optional[int] = 1,  # TODO change later to 20
    n_threads: Optional[int] = 96,  # hard coded to the cluster, change if necessary
    seed: Optional[int] = None,
):
    """CLI to run node2vec."""
    if seed is None:
        seed = random.randint(1, 2 ** 32 - 1)  # noqa: S311
        logger.info(f"random seed given, setting to: {seed}")
    else:
        logger.info(f"random seed set given is: {seed}")

    np.random.seed(seed)
    random.seed(seed)

    # Read graph, first read the triples into a dataframe
    triples_df = pd.read_csv(positive_graph_path, sep=sep)
    # Initialize empty Graph and fill it with the triples from the df
    indra_kg_pos = nx.DiGraph()
    for _, row in tqdm(triples_df[["source", "target"]].iterrows(), total=triples_df.shape[0]):
        # FIXME add double relation for some cases
        indra_kg_pos.add_edge(row["source"], row["target"])

    logger.info("Finished loading the KG")

    # indra_kg_pos = nx.read_edgelist(positive_graph_path, delimiter=sep)

    # Hyperparameters
    # see https://github.com/seffnet/seffnet/blob/master/src/seffnet/optimization.py
    negative: int = 5
    # only use 1 iteration for word2vec, see
    # https://datascience.stackexchange.com/questions/9819/number-of-epochs-in-gensim-word2vec-implementation
    iterations: int = 1
    batch_words: int = 1000  # batch words default value in Gensim is 10,000
    walk_length: int = 127
    # has to be the same as the embedding dimension of the NLP model
    dimensions: int = 768
    # TODO: Discuss if one should use these or just set them to 1, see
    #  "embedding a large graph" https://github.com/VHRanger/nodevectors (replace with inverse if != 1.0)
    p = 1.0
    q = 1.0

    # define HPO function for optuna
    def objective(
        trial: optuna.trial.Trial,
    ) -> float:
        """Run HPO on the link prediction task on the KG, based on a LogReg classifier and the auc score."""
        # Leave out large values for epochs due to runtime constraints
        epochs = trial.suggest_categorical("epochs", [2, 4, 8])  # 8, 16])  # 32, 64, 128, 256])
        # Same for window size
        window_size = trial.suggest_int("window_size", 3, 5)  # 7)

        # train the KGE model
        node2vec_model = Node2Vec(
            n_components=dimensions,
            walklen=walk_length,
            epochs=epochs,
            return_weight=p,
            neighbor_weight=q,
            threads=n_threads,
            keep_walks=True,
            verbose=True,
            w2vparams={
                "window": window_size,
                "negative": negative,
                "iter": iterations,
                "batch_words": batch_words,
            },
        )

        node2vec_model.fit(indra_kg_pos)

        # Save a trained model to a file
        with open(
            os.path.join(KG_HPO_DIR, "node2vec_model_{}.pickle".format(trial.number)), "wb"
        ) as fout:
            pickle.dump(node2vec_model, fout)

        # return the auc score for a LogReg classifier on the link prediction task with negative samples
        return run_link_prediction(
            kg=indra_kg_pos,
            model=node2vec_model,
        )

    # Add mlflow callback to log the HPO in mlflow
    mlflow_callback = MLflowCallback(
        tracking_uri=mlflow_tracking_uri,
        metric_name="ROC AUC Score",
    )

    logger.info("Created the MLflow Callback")

    # Create study and run it
    study = optuna.create_study(
        study_name="Node2vec HPO on INDRA KG",
        storage=f"sqlite:///{MODELS_DIR}/kge_indra_hpo.db",
        direction="maximize",
        load_if_exists=True,
    )

    logger.info("Created the study")

    study.optimize(
        objective,
        n_trials=n_optimization_trials,
        callbacks=[mlflow_callback],
    )

    # Remove all the models except for the best one :)
    for filename in os.listdir(KG_HPO_DIR):
        if filename != "node2vec_model_{}.pickle".format(study.best_trial.number):
            os.remove(os.path.join(KG_HPO_DIR, filename))

    # Load the best model
    with open(
        os.path.join(KG_HPO_DIR, "node2vec_model_{}.pickle".format(study.best_trial.number)), "rb"
    ) as fin:
        best_clf: Node2Vec = pickle.load(fin)

    # Delete the HPO database by default
    if delete_database:
        optuna.delete_study(
            study_name="Node2vec HPO on INDRA KG",
            storage=f"sqlite:///{MODELS_DIR}/kge_indra_hpo.db",
        )

    """Save the embeddings"""
    wv = best_clf.model.wv
    sorted_vocab_items = sorted(wv.vocab.items(), key=lambda item: item[1].count, reverse=True)
    vectors = wv.vectors

    with open(os.path.join(KG_HPO_DIR, "embeddings_best_model.tsv"), "w") as emb_file:
        for word, vocab_ in sorted_vocab_items:
            # Write to vectors file
            embeddings = "\t".join(repr(val) for val in vectors[vocab_.index])
            emb_file.write(f"{word}\t{embeddings}\n")

    """Save the random walks"""
    all_random_walks = best_clf.walks

    with open(os.path.join(KG_HPO_DIR, "random_walks_best_model.tsv"), "w") as random_walk_file:
        for node, random_walks in zip(wv.index2entity, all_random_walks):
            random_walks_str = "\t".join(random_walks)
            random_walk_file.write(f"{node}\t{random_walks_str}\n")


def run_node2vec(
    positive_graph_path: Optional[str] = PRETRAINING_PATH,
    sep: Optional[str] = "\t",
    n_threads: Optional[int] = 96,  # hard coded to the cluster, change if necessary
):
    """Run node2vec with no HPO."""
    # Double check the number of expected embeddings
    triples_df = pd.read_csv(positive_graph_path, sep=sep)
    n_expected_nodes = len(set(triples_df["source"]).union(set(triples_df["target"])))
    logger.info(f"{n_expected_nodes} node embeddings are expected")

    # Use CSRGraph for speedup
    indra_kg_pos = cg.read_edgelist(
        positive_graph_path,
        directed=False,
        sep=sep,
        usecols=["source", "target"],
        header=0,
    )
    logger.info("Finished loading the KG")

    # Hyperparameters
    # see https://github.com/seffnet/seffnet/blob/master/src/seffnet/optimization.py
    negative: int = 5
    # only use 1 iteration for word2vec, see
    # https://datascience.stackexchange.com/questions/9819/number-of-epochs-in-gensim-word2vec-implementation
    iterations: int = 1
    batch_words: int = 10000  # batch words default value in Gensim is 10,000
    walk_length: int = 127
    # has to be the same as the embedding dimension of the NLP model
    dimensions: int = 768
    epochs = 4
    window_size = 3

    # train the KGE model
    node2vec_model = Node2Vec(
        n_components=dimensions,
        walklen=walk_length,
        epochs=epochs,
        # use inverse, see https://github.com/VHRanger/nodevectors/blob/master/nodevectors/node2vec.py#L46
        return_weight=1.0,
        neighbor_weight=1.0,
        threads=n_threads,
        keep_walks=True,
        verbose=True,
        w2vparams={
            "window": window_size,
            "negative": negative,
            "iter": iterations,
            "batch_words": batch_words,
            "min_count": 1,  # do not sort out any words (KG entities)!
        },
    )
    logger.info("Successfully created the model")

    node2vec_model.fit(indra_kg_pos)
    logger.info("Successfully trained the model")

    # Save the trained model to a file
    with open(os.path.join(KG_HPO_DIR, "node2vec_model_no_hpo.pickle"), "wb") as fout:
        pickle.dump(node2vec_model, fout)

    # Save the embeddings
    wv = node2vec_model.model.wv
    sorted_vocab_items = sorted(wv.vocab.items(), key=lambda item: item[1].count, reverse=True)
    logger.info(f"{len(sorted_vocab_items)} embeddings were learned")
    vectors = wv.vectors

    with open(os.path.join(KG_HPO_DIR, "embeddings_best_model.tsv"), "w") as emb_file:
        for word, vocab_ in sorted_vocab_items:
            # Write to vectors file
            embeddings = "\t".join(repr(val) for val in vectors[vocab_.index])
            emb_file.write(f"{word}\t{embeddings}\n")

    logger.info("Successfully saved the embeddings")

    # Save the random walks
    all_random_walks = node2vec_model.walks
    logger.info(f"{len(all_random_walks)} random walks were learned")

    with open(os.path.join(KG_HPO_DIR, "random_walks_best_model.tsv"), "w") as random_walk_file:
        for node, random_walks in zip(wv.index2entity, all_random_walks):
            random_walks_str = "\t".join(random_walks)
            random_walk_file.write(f"{node}\t{random_walks_str}\n")

    logger.info("Successfully saved the walks")


if __name__ == "__main__":
    run_node2vec()



# -*- coding: utf-8 -*-

"""Constants."""

import os

from dotenv import load_dotenv

HERE = os.path.dirname(os.path.realpath(__file__))
SOURCE_DIR = os.path.join(os.path.abspath(os.path.join(HERE, os.pardir)))
PROJECT_DIR = os.path.join(os.path.abspath(os.path.join(HERE, os.pardir)))

# Move to parent folder by os.sep.join(PROJECT_DIR.split(os.sep)[:-1]) to get to the data folder
DATA_DIR = os.path.join(os.sep.join(PROJECT_DIR.split(os.sep)[:-1]), "data")

# Sub-directories of data
RAW_DIR = os.path.join(DATA_DIR, "raw")
INPUT_DIR = os.path.join(DATA_DIR, "input")
OUTPUT_DIR = os.path.join(DATA_DIR, "output")
MISC_DIR = os.path.join(DATA_DIR, "misc")

# Directories for each annotation type
CORRECT_DIR = os.path.join(INPUT_DIR, "correct_incorrect")
DISEASE_DIR = os.path.join(INPUT_DIR, "disease")
LOCATION_DIR = os.path.join(INPUT_DIR, "location")
CELL_LINE_DIR = os.path.join(INPUT_DIR, "cell_line")
SPECIES_DIR = os.path.join(INPUT_DIR, "species")
RELATION_TYPE_DIR = os.path.join(INPUT_DIR, "relation_type")
NDD_DIR = os.path.join(INPUT_DIR, "ndd")

# Path for the pretraining data
PRETRAINING_DIR = os.path.join(INPUT_DIR, "pretraining")
PRETRAINING_PATH = os.path.join(PRETRAINING_DIR, "pretraining_triples.tsv")
PRETRAINING_PREPROCESSED_DF_PATH = os.path.join(PRETRAINING_DIR, "pretraining_preprocessed.pkl")
PRETRAINING_PREPROCESSED_POSITIVE_DF_PATH = os.path.join(
    PRETRAINING_DIR, "pretraining_preprocessed_positive.pkl"
)

# Move to parent folder by os.sep.join(PROJECT_DIR.split(os.sep)[:-1]) to get to the models folder
MODELS_DIR = os.path.join(os.sep.join(PROJECT_DIR.split(os.sep)[:-1]), "models")
NLP_BL_OUTPUT_DIR = os.path.join(MODELS_DIR, "nlp-baseline")
KG_HPO_DIR = os.path.join(MODELS_DIR, "kg-hpo")
KG_BL_OUTPUT_DIR = os.path.join(MODELS_DIR, "kg-baseline")
STONKGS_PRETRAINING_DIR = os.path.join(MODELS_DIR, "stonkgs-pretraining")
STONKGS_PRETRAINING_NO_NSP_DIR = os.path.join(MODELS_DIR, "stonkgs-pretraining-no-nsp")
PRETRAINED_STONKGS_DUMMY_PATH = os.path.join(
    STONKGS_PRETRAINING_DIR, "pretrained-stonkgs-dummy-model"
)
STONKGS_OUTPUT_DIR = os.path.join(MODELS_DIR, "stonkgs")
DEEPSPEED_CONFIG_PATH = os.path.join(MODELS_DIR, "deepspeed_config_zero2.json")

EMBEDDINGS_PATH = os.path.join(KG_HPO_DIR, "embeddings_best_model.tsv")
RANDOM_WALKS_PATH = os.path.join(KG_HPO_DIR, "random_walks_best_model.tsv")

# Move to parent folder by os.sep.join(PROJECT_DIR.split(os.sep)[:-1]) to get to the logs folder
LOG_DIR = os.path.join(os.sep.join(PROJECT_DIR.split(os.sep)[:-1]), "logs")

# Use dotenv to properly load the device-dependent mlflow tracking URI
load_dotenv()
# Load constants from environment variables (set in the .env file)
MLFLOW_TRACKING_URI = os.getenv("MLFLOW_TRACKING_URI")
MLFLOW_FINETUNING_TRACKING_URI = os.getenv("MLFLOW_FINETUNING_TRACKING_URI")
# Load a constant for distinguishing between local and cluster execution (default = True)
LOCAL_EXECUTION = os.getenv("LOCAL_EXECUTION") or "True"

# Directory for visualizations, save it in the notebooks dir for now
NOTEBOOKS_DIR = os.path.join(os.sep.join(PROJECT_DIR.split(os.sep)[:-1]), "notebooks")
VISUALIZATIONS_DIR = os.path.join(NOTEBOOKS_DIR, "visualization")

"""Create directories"""
os.makedirs(DATA_DIR, exist_ok=True)
os.makedirs(MODELS_DIR, exist_ok=True)
os.makedirs(LOG_DIR, exist_ok=True)
os.makedirs(RAW_DIR, exist_ok=True)
os.makedirs(INPUT_DIR, exist_ok=True)
os.makedirs(OUTPUT_DIR, exist_ok=True)
os.makedirs(MISC_DIR, exist_ok=True)
os.makedirs(KG_HPO_DIR, exist_ok=True)
os.makedirs(KG_BL_OUTPUT_DIR, exist_ok=True)
os.makedirs(STONKGS_OUTPUT_DIR, exist_ok=True)
os.makedirs(NLP_BL_OUTPUT_DIR, exist_ok=True)
os.makedirs(VISUALIZATIONS_DIR, exist_ok=True)
os.makedirs(NOTEBOOKS_DIR, exist_ok=True)

os.makedirs(PRETRAINING_DIR, exist_ok=True)
os.makedirs(DISEASE_DIR, exist_ok=True)
os.makedirs(LOCATION_DIR, exist_ok=True)
os.makedirs(CELL_LINE_DIR, exist_ok=True)
os.makedirs(SPECIES_DIR, exist_ok=True)
os.makedirs(RELATION_TYPE_DIR, exist_ok=True)

# Specify the raw (complete) INDRA json file
INDRA_RAW_JSON = os.path.join(RAW_DIR, "raw_statements.json")
# Download from https://emmaa.s3.amazonaws.com/assembled/covid19/statements_2021-03-08-18-24-29.gz
DUMMY_EXAMPLE_INDRA = os.path.join(RAW_DIR, "statements_2021-01-30-17-21-54.json")
# Can be created by running python -m src.stonkgs.data.indra
DUMMY_EXAMPLE_TRIPLES = os.path.join(LOCATION_DIR, "location.tsv")

# Specify the (huggingface) language model that is used in this project
NLP_MODEL_TYPE = "dmis-lab/biobert-v1.1"

# Specify the vocab file of the language model that is used in this project
# (the file can be obtained here: https://huggingface.co/dmis-lab/biobert-v1.1/tree/main)
VOCAB_FILE = os.path.join(MISC_DIR, "biobert_vocab.txt")



# -*- coding: utf-8 -*-
"""
KG baseline model on the fine-tuning classification task, assuming the model embeddings are pre-trained.

Run with:
python -m src.stonkgs.models.kg_baseline_model
"""

import logging
import os
from collections import Counter
from typing import Dict, List, Optional

import click  # type: ignore
import mlflow
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from sklearn.metrics import f1_score
from sklearn.model_selection import KFold, StratifiedShuffleSplit

from stonkgs.constants import (
    CELL_LINE_DIR,
    CORRECT_DIR,
    DISEASE_DIR,
    EMBEDDINGS_PATH,
    KG_BL_OUTPUT_DIR,
    LOCATION_DIR,
    MLFLOW_FINETUNING_TRACKING_URI,
    RANDOM_WALKS_PATH,
    RELATION_TYPE_DIR,
    SPECIES_DIR,
)

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
# Disable alembic info
logging.getLogger("alembic").setLevel(logging.WARNING)


class KGEClassificationModel(pl.LightningModule):
    """KGE baseline model."""

    def __init__(
        self,
        num_classes,
        class_weights,
        d_in: int = 768,
        lr: float = 1e-3,
    ):
        """Initialize the components of the KGE based classification model.

        :param num_classes: number of classes
        :param class_weights: class weights
        :param d_in: dimensions
        :param lr: learning rate

        The model consists of
        1) "Max-Pooling" (embedding-dimension-wise max)
        2) Dropout
        3) Linear layer (d_in x num_classes)
        4) Softmax
        (Not part of the model, but of the class: class_weights for the cross_entropy function)
        """
        super(KGEClassificationModel, self).__init__()

        # Model architecture
        self.pooling = torch.max
        self.dropout = torch.nn.Dropout(0.1)
        self.linear = torch.nn.Linear(d_in, num_classes)
        self.softmax = torch.nn.Softmax(dim=1)

        # Other class-specific parameters
        # Class weights for CE loss
        self.class_weights = torch.tensor(class_weights)
        # Learning rate
        self.lr = lr

        # Add an attribute to save the predictions later on (will be overridden anyways)
        self.predicted_labels = torch.empty(size=(1,))

        # Log the additional parameters
        self.log_dict({"num_classes": num_classes, "class_weights": class_weights, "lr": lr})

    def forward(self, x):
        """Perform forward pass consisting of pooling (dimension-wise max), and a linear layer followed by softmax.

        :param x: embedding sequences (random walk embeddings) for the given triples
        :return: class probabilities for the given triples
        """
        h_pooled = self.pooling(x, dim=1).values
        dropout_output = self.dropout(h_pooled)
        linear_output = self.linear(dropout_output)
        y_pred = self.softmax(linear_output)
        # Note that the forward pass returns class probabilities.
        return y_pred

    def configure_optimizers(self):
        """Configure optimizer and scheduler."""
        optimizer = torch.optim.AdamW(self.parameters(), lr=self.lr)
        return optimizer

    def training_step(self, batch, batch_idx):
        """Perform one training step on one batch using CE loss."""
        train_inputs, train_labels = batch
        train_outputs = self.forward(train_inputs)
        loss_fct = torch.nn.CrossEntropyLoss(reduction="mean", weight=self.class_weights)
        loss = loss_fct(train_outputs, train_labels)

        # Log loss at each training step
        self.log("loss", loss.item(), on_step=True)

        return loss

    def test_step(self, batch, batch_nb):
        """Perform one test step on a given batch and return the weighted-averaged f1 score over all batches."""
        test_inputs, test_labels = batch
        test_class_probs = self.forward(test_inputs)
        # Class probabilities to class labels
        test_predictions = torch.argmax(test_class_probs, dim=1)
        # logger.info(f'Predicted labels: {test_predictions}')

        return {"labels": test_labels, "predictions": test_predictions}

    def test_epoch_end(self, outputs):
        """Return average and std weighted-averaged f1-score over all batches."""
        # Get the weighted-averaged f1-score
        all_labels = torch.cat([x["labels"] for x in outputs])
        all_predictions = torch.cat([x["predictions"] for x in outputs])
        test_f1 = f1_score(all_labels, all_predictions, average="weighted")

        # Log the final f1 score
        self.log("f1_score_weighted", test_f1)

        # Use this (weird) fix to save the predicted labels
        self.predicted_labels = all_predictions

        return {"test_f1": test_f1}


class INDRAEntityDataset(torch.utils.data.Dataset):
    """Custom dataset class for INDRA data."""

    def __init__(self, embedding_dict, random_walk_dict, sources, targets, labels, max_len=254):
        """Initialize INDRA Dataset based on random walk embeddings for 2 nodes in each triple."""
        self.max_length = max_len
        # Two entities (source, target) of each triple
        self.sources = sources
        self.targets = targets
        # Initialize dictionary of node name -> embedding vector
        self.embedding_dict = embedding_dict
        # Add the null vector to the embedding dict for "out-of-vocabulary" nodes
        self.embedding_dict[-1] = np.zeros(np.shape(next(iter(self.embedding_dict.values()))))
        # Initialize dictionary of node name -> random walk node names
        self.random_walk_dict = random_walk_dict
        # Get the embedding sequences for each triple
        self.embeddings = self.get_embeddings()
        # Assumes that the labels are numerically encoded
        self.labels = labels

    def __getitem__(self, idx):
        """Get embeddings and labels for given indices."""
        # Get embeddings (of random walk sequences of source + target) for given indices
        item = torch.tensor(self.embeddings[idx, :, :], dtype=torch.float)
        # Get labels for given indices
        labels = torch.tensor(self.labels[idx], dtype=torch.long)
        return item, labels

    def __len__(self):
        """Return the length of the dataset."""
        return len(self.labels)

    def get_embeddings(self):
        """Get the embedding sequences for each triple in the dataset (node emb from sources + targets random walks).

        :return: embedding sequences for each triple in the dataset
        """
        # Number of total examples in the dataset
        number_of_triples = len(self.sources)
        # Get the embedding dimension by accessing a random element
        embedding_dim = len(next(iter(self.embedding_dict.values())))

        # Initialize the embedding array of dimension n x random_walk_length x embedding_dim
        embeddings = np.empty((number_of_triples, self.max_length, embedding_dim))

        # 1. Iterate through all triples: Get random walks for sources and targets using random_walk_dict
        for idx, (source, target) in enumerate(zip(self.sources, self.targets)):
            # Get the random walk sequences
            random_walk_source = self.random_walk_dict[source]
            random_walk_target = self.random_walk_dict[target]

            # 2. Concatenate and the random walks
            # The total random walk has the length max_length. Therefore its split half into the random walk of source
            # and half target.
            random_walk = random_walk_source.tolist() + random_walk_target.tolist()  # noqa: N400
            # 3. Get embeddings for each node using embedding_dict stated by its index in each random walk
            embeds_random_walk = np.stack(
                [self.embedding_dict[node] for node in random_walk], axis=0
            )
            # The final embedding sequence for a given triple has the dimension max_length x embedding_dim
            embeddings[idx, :, :] = embeds_random_walk

        return embeddings


def prepare_df(embedding_path: str, sep: str = "\t") -> Dict[str, List[str]]:
    """Prepare dataframe to node->embeddings/random walks."""
    # Load embeddings
    df = pd.read_csv(
        embedding_path,
        sep=sep,
        header=None,
        index_col=0,
    )
    # Node id -> embeddings
    return {index: row.values for index, row in df.iterrows()}


def get_train_test_splits(
    data: pd.DataFrame,
    label_column_name: str = "class",
    random_seed: int = 42,
    n_splits: int = 5,
    max_dataset_size: int = 100000,
) -> List:
    """Return deterministic train/test indices for n_splits based on the fine-tuning dataset that is passed."""
    # Leave out the label in the dataset
    data_no_labels = data.drop(label_column_name, axis=1)
    labels = data[label_column_name]

    # Cut the dataset down to max_dataset_size (deterministically!) using StratifiedShuffleSplit if needed:
    # (this is not an actual train/test split, this is just for getting a dataset of size max_dataset_size in a
    # stratified and deterministic manner)
    if len(data) > max_dataset_size:
        splitter = StratifiedShuffleSplit(
            n_splits=1,
            train_size=max_dataset_size,
            random_state=random_seed,
        )
        for train_index, _ in splitter.split(data_no_labels, labels):
            data_no_labels = data_no_labels.iloc[train_index, :].reset_index(drop=True)
            labels = labels.iloc[train_index].reset_index(drop=True)

    # Generate the actual train/test splits here:
    # Implement non-stratified train/test splits with no validation split
    # It is shuffled deterministically (determined by random_seed)
    skf = KFold(n_splits=n_splits, random_state=random_seed, shuffle=True)

    # Return a list of dictionaries for train and test indices
    return [
        {"train_idx": train_idx, "test_idx": test_idx}
        for train_idx, test_idx in skf.split(data_no_labels, labels)
    ]


def run_kg_baseline_classification_cv(
    triples_path: str,
    embedding_path: str = EMBEDDINGS_PATH,
    random_walks_path: str = RANDOM_WALKS_PATH,
    logging_uri_mlflow: Optional[str] = MLFLOW_FINETUNING_TRACKING_URI,
    n_splits: int = 5,
    epochs: int = 100,
    train_batch_size: int = 8,
    test_batch_size: int = 64,
    lr: float = 1e-3,
    label_column_name: str = "class",
    log_steps: int = 500,
    task_name: str = "",
    max_dataset_size: int = 100000,
) -> Dict[str, float]:
    """Run KG baseline classification."""
    # Step 1. load the tsv file with the annotation types you want to test and make the splits
    triples_df = pd.read_csv(
        triples_path,
        sep="\t",
        usecols=[
            "source",
            "target",
            label_column_name,
        ],
    )

    # Prepare embeddings and random walks
    embeddings_dict = prepare_df(embedding_path)
    random_walks_dict = prepare_df(random_walks_path)

    # Filter out any triples that contain a node that is not in the embeddings_dict
    original_length = len(triples_df)
    triples_df = triples_df[
        triples_df["source"].isin(embeddings_dict.keys())
        & triples_df["target"].isin(embeddings_dict.keys())
    ].reset_index(drop=True)
    new_length = len(triples_df)
    logger.info(
        f"{original_length - new_length} out of {original_length} triples are left out because they contain "
        f"nodes which are not present in the pre-training data"
    )

    # Numerically encode labels
    unique_tags = set(label for label in triples_df[label_column_name])
    tag2id = {label: number for number, label in enumerate(unique_tags)}
    id2tag = {value: key for key, value in tag2id.items()}

    # Get labels
    labels = pd.Series([int(tag2id[label]) for label in triples_df[label_column_name]])

    # Get the train/test split indices
    train_test_splits = get_train_test_splits(
        triples_df,
        n_splits=n_splits,
        label_column_name=label_column_name,
        max_dataset_size=max_dataset_size,
    )

    # Initialize f1-scores
    f1_scores = []

    # Initialize INDRA for KG baseline dataset
    kg_embeds = INDRAEntityDataset(
        embeddings_dict,
        random_walks_dict,
        triples_df["source"],
        triples_df["target"],
        labels,
    )

    mlflow.set_tracking_uri(logging_uri_mlflow)
    mlflow.set_experiment("KG Baseline for STonKGs")

    mlflow.pytorch.autolog()

    # Initialize a dataframe for all the predicted labels
    result_df = pd.DataFrame()

    # Train and test the model in a cv setting
    for idx, indices in enumerate(train_test_splits):
        # Sample elements randomly from a given list of ids, no replacement
        train_subsampler = torch.utils.data.SubsetRandomSampler(indices["train_idx"])
        test_subsampler = torch.utils.data.SubsetRandomSampler(indices["test_idx"])
        # CE class weights for the model based on training data class distribution,
        # based on the class counts (Inverse Number of Samples, INS)
        weights = [
            1
            / len(
                [
                    i
                    for i in triples_df.iloc[indices["train_idx"], :][label_column_name]
                    # note that we only employ train idx
                    if i == id2tag[id_num]
                ]
            )
            for id_num in range(len(unique_tags))
        ]

        # Define data loaders for training and testing data in this fold
        trainloader = torch.utils.data.DataLoader(
            kg_embeds,
            batch_size=train_batch_size,
            sampler=train_subsampler,
        )
        testloader = torch.utils.data.DataLoader(
            kg_embeds,
            batch_size=test_batch_size,
            sampler=test_subsampler,
        )

        model = KGEClassificationModel(
            num_classes=len(triples_df[label_column_name].unique()),
            class_weights=weights,
            lr=lr,
        )

        # Initialize pytorch lighting Trainer for the KG baseline model
        trainer = pl.Trainer(
            max_epochs=epochs,
            log_every_n_steps=log_steps,
        )

        # Train and predict in a separate run for each split
        with mlflow.start_run():
            # Fit on training split
            trainer.fit(model, train_dataloader=trainloader)
            # Predict on test split
            test_results = trainer.test(model, test_dataloaders=testloader)

            # Save the predicted + true labels
            partial_result_df = pd.DataFrame(
                {
                    "split": idx,
                    "index": indices["test_idx"].tolist(),
                    "predicted_label": model.predicted_labels.tolist(),
                    "true_label": labels[indices["test_idx"]].tolist(),
                },
            )
            result_df = result_df.append(
                partial_result_df,
                ignore_index=True,
            )

            # Log some details about the datasets used in training and testing
            mlflow.log_param("label dict", str(tag2id))
            mlflow.log_param("training dataset size", str(len(trainloader.dataset)))
            mlflow.log_param("training class dist", str(Counter(trainloader.dataset.labels)))
            mlflow.log_param("test dataset size", str(len(testloader.dataset)))
            mlflow.log_param("test class dist", str(Counter(testloader.dataset.labels)))

        # Append f1 score per split based on the weighted average
        f1_scores.append(test_results[0]["test_f1"])

    # Map the labels in the result df back to their original names
    result_df = result_df.replace({"predicted_label": id2tag, "true_label": id2tag})
    # Save the result_df
    result_df.to_csv(
        os.path.join(KG_BL_OUTPUT_DIR, "predicted_labels_kg_" + task_name + "df.tsv"),
        index=False,
        sep="\t",
    )

    # Save the last model
    trainer.save_checkpoint(os.path.join(KG_BL_OUTPUT_DIR, "kg_baseline.ckpt"))

    # Log the mean and std f1 score from the cross validation procedure to mlflow
    with mlflow.start_run():
        mlflow.log_metric("f1_score_mean", np.mean(f1_scores))
        mlflow.log_metric("f1_score_std", np.std(f1_scores))

        # Log the task name as well
        mlflow.log_param("task name", task_name)

        # Also log how many triples were left out
        mlflow.log_param("original no. of triples", original_length)
        mlflow.log_param("no. of left out triples", original_length - new_length)

    # Log mean and std f1-scores from the cross validation procedure (average and std across all splits) to the
    # standard logger
    logger.info(f"Mean f1-score: {np.mean(f1_scores)}")
    logger.info(f"Std f1-score: {np.std(f1_scores)}")

    # Return the final f1 score mean and the std for all CV folds
    return {"f1_score_mean": np.mean(f1_scores), "f1_score_std": float(np.std(f1_scores))}


@click.command()
@click.option("-e", "--epochs", default=5, help="Number of epochs", type=int)
@click.option("--lr", default=1e-4, help="Learning rate", type=float)
@click.option(
    "--logging_dir",
    default=MLFLOW_FINETUNING_TRACKING_URI,
    help="Mlflow logging/tracking URI",
    type=str,
)
@click.option("--log_steps", default=500, help="Number of steps between each log", type=int)
@click.option("--batch_size", default=16, help="Batch size", type=int)
@click.option(
    "--max_dataset_size",
    default=100000,
    help="Maximum dataset size of the fine-tuning datasets",
    type=int,
)
def run_all_fine_tuning_tasks(
    epochs: int = 5,
    log_steps: int = 500,
    lr: float = 1e-3,
    logging_dir: Optional[str] = MLFLOW_FINETUNING_TRACKING_URI,
    batch_size: int = 16,
    max_dataset_size: int = 100000,  # effectively removes the max dataset size restriction
):
    """Run all fine-tuning tasks at once."""
    # Specify all directories and file names
    directories = [
        CELL_LINE_DIR,
        CORRECT_DIR,
        CORRECT_DIR,
        DISEASE_DIR,
        LOCATION_DIR,
        SPECIES_DIR,
        RELATION_TYPE_DIR,
        RELATION_TYPE_DIR,
    ]
    file_names = [
        "cell_line_no_duplicates.tsv",
        "correct_incorrect_binary_no_duplicates.tsv",
        "correct_incorrect_multiclass_no_duplicates.tsv",
        "disease_no_duplicates.tsv",
        "location_no_duplicates.tsv",
        "species_no_duplicates.tsv",
        "relation_type_no_duplicates.tsv",
        "relation_type_no_duplicates.tsv",
    ]
    task_names = [
        "cell_line",
        "correct_binary",
        "correct_multiclass",
        "disease",
        "location",
        "species",
        "interaction",
        "polarity",
    ]
    # Specify the column names of the target variable
    column_names = ["class"] * 6 + ["interaction"] + ["polarity"]

    for directory, file, column_name, task_name in zip(
        directories,
        file_names,
        column_names,
        task_names,
    ):
        # Run each of the six classification tasks
        run_kg_baseline_classification_cv(
            triples_path=os.path.join(directory, file),
            label_column_name=column_name,
            logging_uri_mlflow=logging_dir,
            epochs=epochs,
            lr=lr,
            log_steps=log_steps,
            train_batch_size=batch_size,
            task_name=task_name,
            max_dataset_size=max_dataset_size,
        )
        logger.info(f"Finished the {task_name} task")


if __name__ == "__main__":
    # Run all fine-tuning classification tasks at once
    run_all_fine_tuning_tasks()



# -*- coding: utf-8 -*-

"""Checks if all entities in the fine-tuning datasets are contained in the pre-training data.

Run with:
python -m src.stonkgs.data.indra_check_overlaps
"""

import logging
import os
from typing import Dict, Set

import pandas as pd

from stonkgs.constants import (
    PRETRAINING_PATH,
    CELL_LINE_DIR,
    DISEASE_DIR,
    LOCATION_DIR,
    RELATION_TYPE_DIR,
    SPECIES_DIR,
)

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def load_entities(
    df_path: str,
) -> Set[str]:
    """Returns all entities in a given dataset."""
    df = pd.read_csv(df_path, sep="\t")
    logger.info(f"Number of triples: {len(df)}")
    all_entities = set(df["source"]).union(set(df["target"]))
    logger.info(f"Successfully loaded {len(all_entities)} many entities \n")

    return all_entities


def find_missing_entities(
    pre_training_entities: Set[str],
    fine_tuning_entities_dict: Dict[str, Set[str]],
):
    """Identifies the entities (if any) present in the fine tuning BUT NOT in the pretraining dataset."""
    for name, fine_tuning_entities in fine_tuning_entities_dict.items():
        new_entities = fine_tuning_entities.difference(pre_training_entities)
        logger.info(
            f"For {name}, there are {len(new_entities)} many entities present in the fine-tuning set that are "
            f"not in the pre-training data"
        )
        # logger.info(new_entities)
        logger.info("\n")


def load_text(
    df_path: str,
) -> Set[str]:
    """Returns all text evidences in a given dataset."""
    df = pd.read_csv(df_path, sep="\t")
    all_text = set(df["evidence"])
    logger.info(f"Successfully loaded {len(all_text)} many text evidences")

    return all_text


def find_information_leakage(
    pre_training_text_evidences: Set[str],
    fine_tuning_evidences_dict: Dict[str, Set[str]],
):
    """Identifies how many text evidences from the fine-tuning sets are present in the pretraining dataset."""
    for name, fine_tuning_ev in fine_tuning_evidences_dict.items():
        new_entities = fine_tuning_ev.intersection(pre_training_text_evidences)
        logger.info(
            f"For {name}, there are {len(new_entities)} out of {len(fine_tuning_ev)} many entities that are "
            f"also present in the pre-training data"
        )
        # logger.info(new_entities)


if __name__ == "__main__":
    # 1. ENTITY OVERLAP
    # Get all the pre-training and fine-tuning datasets
    pre_training_ents = load_entities(PRETRAINING_PATH)

    # Iterate through all the fine-tuning stuff for ENTITIES
    paths = [CELL_LINE_DIR, DISEASE_DIR, LOCATION_DIR, RELATION_TYPE_DIR, SPECIES_DIR]
    names = [
        "cell_line_filtered",
        "disease_filtered",
        "location_filtered",
        "relation_type",
        "species_filtered",
    ]

    fine_tuning_dict = dict()
    for path, annot_name in zip(paths, names):
        fine_tuning_dict[annot_name] = load_entities(os.path.join(path, annot_name + ".tsv"))

    # Get an estimate of the entities that are not in the pretraining data
    find_missing_entities(
        pre_training_entities=pre_training_ents,
        fine_tuning_entities_dict=fine_tuning_dict,
    )

    # 2. INFORMATION LEAKAGE
    # Make sure that there is no information leakage, i.e. text evidence that is in both pre-training and one of
    # the fine-tuning datasets
    # Get the pre-training texts
    pre_training_texts = load_text(PRETRAINING_PATH)

    # Iterate through all the fine-tuning stuff for EVIDENCES
    fine_tuning_dict_evidences = {}
    for path, annot_name in zip(paths, names):
        fine_tuning_dict_evidences[annot_name] = load_text(os.path.join(path, annot_name + ".tsv"))

    # Get an estimate of the entities that are not in the pretraining data
    find_information_leakage(
        pre_training_text_evidences=pre_training_texts,
        fine_tuning_evidences_dict=fine_tuning_dict_evidences,
    )



# -*- coding: utf-8 -*-

"""NLP baseline model on the fine-tuning classification task, assuming the model embeddings are pre-trained."""

import logging
import os
from collections import Counter
from typing import Dict, List, Optional

import click  # type: ignore
import mlflow
import numpy as np
import pandas as pd
import torch
from sklearn.metrics import f1_score
from sklearn.model_selection import KFold, StratifiedShuffleSplit
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    Trainer,
    TrainingArguments,
)

from stonkgs.constants import (
    CELL_LINE_DIR,
    CORRECT_DIR,
    DEEPSPEED_CONFIG_PATH,
    DISEASE_DIR,
    EMBEDDINGS_PATH,
    LOCATION_DIR,
    MLFLOW_FINETUNING_TRACKING_URI,
    NLP_BL_OUTPUT_DIR,
    NLP_MODEL_TYPE,
    RELATION_TYPE_DIR,
    SPECIES_DIR,
    STONKGS_OUTPUT_DIR,
)
from stonkgs.data.indra_for_pretraining import prepare_df

# Initialize logger
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
# Disable alembic info
logging.getLogger("alembic").setLevel(logging.WARNING)


class INDRAEvidenceDataset(torch.utils.data.Dataset):
    """Custom Dataset class for INDRA data."""

    def __init__(self, encodings, labels):
        """Initialize INDRA Dataset based on token embeddings for each text evidence."""
        # Assumes that the labels are numerically encoded
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        """Return data entries (text evidences) for given indices."""
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        """Return the length of the dataset."""
        return len(self.labels)


def get_train_test_splits(
    data: pd.DataFrame,
    max_dataset_size: int = 100000,
    label_column_name: str = "class",
    random_seed: int = 42,
    n_splits: int = 5,
) -> List:
    """Return deterministic train/test indices for n_splits based on the fine-tuning dataset that is passed."""
    # Leave out the label in the dataset
    data_no_labels = data.drop(label_column_name, axis=1)
    labels = data[label_column_name]

    # Cut the dataset down to max_dataset_size (deterministically!) using StratifiedShuffleSplit if needed:
    # (this is not an actual train/test split, this is just for getting a dataset of size max_dataset_size in a
    # stratified and deterministic manner)
    if len(data) > max_dataset_size:
        splitter = StratifiedShuffleSplit(
            n_splits=1,
            train_size=max_dataset_size,
            random_state=random_seed,
        )
        for train_index, _ in splitter.split(data_no_labels, labels):
            data_no_labels = data_no_labels.iloc[train_index, :].reset_index(drop=True)
            labels = labels.iloc[train_index].reset_index(drop=True)

    # Generate the actual train/test splits here:
    # Implement non-stratified train/test splits with no validation split
    # It is shuffled deterministically (determined by random_seed)
    skf = KFold(n_splits=n_splits, random_state=random_seed, shuffle=True)

    return [
        {"train_idx": train_idx, "test_idx": test_idx}
        for train_idx, test_idx in skf.split(data_no_labels, labels)
    ]


def run_nlp_baseline_classification_cv(
    train_data_path: str,
    sep: Optional[str] = "\t",
    model_type: str = NLP_MODEL_TYPE,
    output_dir: Optional[str] = NLP_BL_OUTPUT_DIR,
    logging_uri_mlflow: Optional[str] = MLFLOW_FINETUNING_TRACKING_URI,
    label_column_name: str = "class",
    text_data_column_name: str = "evidence",
    epochs: Optional[int] = 10,
    log_steps: int = 500,
    lr: float = 5e-5,
    batch_size: int = 16,
    gradient_accumulation: int = 1,
    task_name: str = "",
    embedding_path: str = EMBEDDINGS_PATH,
    deepspeed: bool = True,
    max_dataset_size: int = 100000,
) -> Dict:
    """Run cross-validation for the sequence classification task."""
    # Get data splits
    indra_data = pd.read_csv(train_data_path, sep=sep)
    # TODO: leave it out later on?
    # Filter out any triples that contain a node that is not in the embeddings_dict
    embeddings_dict = prepare_df(embedding_path)
    original_length = len(indra_data)
    indra_data = indra_data[
        indra_data["source"].isin(embeddings_dict.keys())
        & indra_data["target"].isin(embeddings_dict.keys())
    ].reset_index(drop=True)
    new_length = len(indra_data)
    logger.info(
        f"{original_length - new_length} out of {original_length} triples are left out because they contain "
        f"nodes which are not present in the pre-training data"
    )

    train_test_splits = get_train_test_splits(
        indra_data,
        label_column_name=label_column_name,
        max_dataset_size=max_dataset_size,
    )

    # Get text evidences and labels
    evidences_text, labels_str = indra_data[text_data_column_name], indra_data[label_column_name]
    # Numerically encode labels
    unique_tags = set(label for label in labels_str)
    tag2id = {label: number for number, label in enumerate(unique_tags)}
    id2tag = {value: key for key, value in tag2id.items()}
    labels = pd.Series([int(tag2id[label]) for label in labels_str])

    # Initialize the f1-score
    f1_scores = []

    # End previous run
    mlflow.end_run()
    # Initialize mlflow run, set tracking URI to use the same experiment for all runs,
    # so that one can compare them
    mlflow.set_tracking_uri(logging_uri_mlflow)
    mlflow.set_experiment("NLP Baseline for STonKGs")

    # Start a parent run so that all CV splits are tracked as nested runs
    # mlflow.start_run(run_name='Parent Run')

    # Initialize a dataframe for all the predicted labels
    result_df = pd.DataFrame()

    for idx, indices in enumerate(train_test_splits):
        # Initialize tokenizer and model
        tokenizer = AutoTokenizer.from_pretrained(model_type)
        model = AutoModelForSequenceClassification.from_pretrained(
            model_type, num_labels=len(unique_tags)
        )

        # Encode all text evidences, pad and truncate to max_seq_len
        train_evidences = tokenizer(
            evidences_text[indices["train_idx"]].tolist(), truncation=True, padding=True
        )
        test_evidences = tokenizer(
            evidences_text[indices["test_idx"]].tolist(), truncation=True, padding=True
        )
        train_labels = labels[indices["train_idx"]].tolist()
        test_labels = labels[indices["test_idx"]].tolist()
        train_dataset = INDRAEvidenceDataset(encodings=train_evidences, labels=train_labels)
        test_dataset = INDRAEvidenceDataset(encodings=test_evidences, labels=test_labels)

        # Note that due to the randomization in the batches, the training/evaluation is slightly
        # different every time
        training_args = TrainingArguments(
            # label_names
            output_dir=output_dir,
            num_train_epochs=epochs,  # total number of training epochs
            logging_steps=log_steps,
            learning_rate=lr,
            # Use deepspeed with a specified config file for speedup
            deepspeed=DEEPSPEED_CONFIG_PATH if deepspeed else None,
            report_to=["mlflow"],  # log via mlflow
            do_train=True,
            do_predict=True,
            per_device_train_batch_size=batch_size,
            gradient_accumulation_steps=gradient_accumulation,
        )

        # Initialize Trainer based on the training dataset
        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
        )
        # Train
        trainer.train()

        # Log some details about the datasets used in training and testing
        mlflow.log_param("label dict", str(tag2id))
        mlflow.log_param("training dataset size", str(len(train_labels)))
        mlflow.log_param("training class dist", str(Counter(train_labels)))
        mlflow.log_param("test dataset size", str(len(test_labels)))
        mlflow.log_param("test class dist", str(Counter(test_labels)))

        # Make predictions for the test dataset
        predictions = trainer.predict(test_dataset=test_dataset).predictions
        predicted_labels = np.argmax(predictions, axis=1)
        logger.info(f"Predicted labels: {predicted_labels}")

        # Save the predicted + true labels
        partial_result_df = pd.DataFrame(
            {
                "split": idx,
                "index": indices["test_idx"].tolist(),
                "predicted_label": predicted_labels.tolist(),
                "true_label": test_labels,
                "evidence": evidences_text[indices["test_idx"]].tolist(),
            },
        )
        result_df = result_df.append(
            partial_result_df,
            ignore_index=True,
        )

        # Use weighted average
        f1_sc = f1_score(test_labels, predicted_labels, average="weighted")
        f1_scores.append(f1_sc)

        # Log the final f1 score of the split
        mlflow.log_metric("f1_score_weighted", f1_sc)

    # Log mean and std f1-scores from the cross validation procedure (average and std across all splits) to the
    # standard logger
    logger.info(f"Mean f1-score: {np.mean(f1_scores)}")
    logger.info(f"Std f1-score: {np.std(f1_scores)}")

    # Map the labels in the result df back to their original names
    result_df = result_df.replace({"predicted_label": id2tag, "true_label": id2tag})
    # Save the result_df
    result_df.to_csv(
        os.path.join(NLP_BL_OUTPUT_DIR, "predicted_labels_nlp_" + task_name + "df.tsv"),
        index=False,
        sep="\t",
    )

    # Save the last model
    trainer.save_model(output_dir=NLP_BL_OUTPUT_DIR)

    # End the previous run
    mlflow.end_run()

    # Log the mean and std f1 score from the cross validation procedure to mlflow
    with mlflow.start_run():
        # Log the task name as well
        mlflow.log_param("task name", task_name)
        mlflow.log_metric("f1_score_mean", np.mean(f1_scores))
        mlflow.log_metric("f1_score_std", np.std(f1_scores))

    # End parent run
    # mlflow.end_run()

    return {"f1_score_mean": np.mean(f1_scores), "f1_score_std": np.std(f1_scores)}


@click.command()
@click.option("-e", "--epochs", default=5, help="Number of epochs", type=int)
@click.option("--lr", default=5e-5, help="Learning rate", type=float)
@click.option(
    "--logging_dir",
    default=MLFLOW_FINETUNING_TRACKING_URI,
    help="Mlflow logging/tracking URI",
    type=str,
)
@click.option("--log_steps", default=500, help="Number of steps between each log", type=int)
@click.option("--output_dir", default=STONKGS_OUTPUT_DIR, help="Output directory", type=str)
@click.option("--batch_size", default=8, help="Batch size used in fine-tuning", type=int)
@click.option(
    "--gradient_accumulation_steps", default=1, help="Gradient accumulation steps", type=int
)
@click.option("--deepspeed", default=True, help="Whether to use deepspeed or not", type=bool)
@click.option(
    "--max_dataset_size",
    default=100000,
    help="Maximum dataset size of the fine-tuning datasets",
    type=int,
)
@click.option("--local_rank", default=-1, help="THIS PARAMETER IS IGNORED", type=int)
def run_all_fine_tuning_tasks(
    epochs: int = 5,
    log_steps: int = 500,
    lr: float = 5e-5,
    output_dir: str = STONKGS_OUTPUT_DIR,
    logging_dir: Optional[str] = MLFLOW_FINETUNING_TRACKING_URI,
    batch_size: int = 8,
    gradient_accumulation_steps: int = 1,
    deepspeed: bool = True,
    max_dataset_size: int = 100000,  # effectively removes the max dataset size restriction
    local_rank: int = -1,
):
    """Run all fine-tuning tasks at once."""
    # Specify all directories and file names
    directories = [
        CELL_LINE_DIR,
        CORRECT_DIR,
        CORRECT_DIR,
        DISEASE_DIR,
        LOCATION_DIR,
        SPECIES_DIR,
        RELATION_TYPE_DIR,
        RELATION_TYPE_DIR,
    ]
    file_names = [
        "cell_line_no_duplicates.tsv",
        "correct_incorrect_binary_no_duplicates.tsv",
        "correct_incorrect_multiclass_no_duplicates.tsv",
        "disease_no_duplicates.tsv",
        "location_no_duplicates.tsv",
        "species_no_duplicates.tsv",
        "relation_type_no_duplicates.tsv",
        "relation_type_no_duplicates.tsv",
    ]
    task_names = [
        "cell_line",
        "correct_binary",
        "correct_multiclass",
        "disease",
        "location",
        "species",
        "interaction",
        "polarity",
    ]
    # Specify the column names of the target variable
    column_names = ["class"] * 6 + ["interaction"] + ["polarity"]

    for directory, file, column_name, task_name in zip(
        directories,
        file_names,
        column_names,
        task_names,
    ):
        # Run each of the six classification tasks
        run_nlp_baseline_classification_cv(
            train_data_path=os.path.join(directory, file),
            output_dir=output_dir,
            logging_uri_mlflow=logging_dir,
            epochs=epochs,
            log_steps=log_steps,
            lr=lr,
            batch_size=batch_size,
            gradient_accumulation=gradient_accumulation_steps,
            label_column_name=column_name,
            task_name=task_name,
            deepspeed=deepspeed,
            max_dataset_size=max_dataset_size,
        )
        logger.info(f"Finished the {task_name} task")


if __name__ == "__main__":
    # Set the huggingface environment variable for tokenizer parallelism to false
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    # Run all classification tasks
    run_all_fine_tuning_tasks()



# -*- coding: utf-8 -*-

"""Script for running the pre-training procedure of STonKGs."""

import logging
import os
from typing import Optional

import click  # type: ignore
import mlflow
from accelerate import Accelerator
from datasets import Dataset, load_dataset, total_allocated_bytes
from transformers import (
    Trainer,
    TrainingArguments,
)
from transformers.trainer_utils import get_last_checkpoint

from stonkgs.constants import (
    DEEPSPEED_CONFIG_PATH,
    MLFLOW_TRACKING_URI,
    NLP_MODEL_TYPE,
    PRETRAINING_PREPROCESSED_POSITIVE_DF_PATH,
    STONKGS_PRETRAINING_NO_NSP_DIR,
)
from stonkgs.models.stonkgs_model import STonKGsForPreTraining

# Initialize logger
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
# Disable alembic info
logging.getLogger("alembic").setLevel(logging.WARNING)


def _load_pre_training_data(
    pretraining_preprocessed_path: str = PRETRAINING_PREPROCESSED_POSITIVE_DF_PATH,
    dataset_format: str = "torch",
) -> Dataset:
    """Create a pytorch dataset based on a preprocessed dataframe for the pretraining dataset."""
    # Load the pickled dataframe using the load_dataset function so that it can be cached
    pretraining_dataset = load_dataset(
        "pandas", data_files=pretraining_preprocessed_path, split="train"
    )

    # Do not put the dataset on the GPU even if possible, it is only stealing GPU space, use the dataloader instead
    # Putting it on the GPU might only be worth it if 4+ GPUs are used
    pretraining_dataset.set_format(dataset_format)
    logger.info("Finished loading the pretraining dataset")

    return pretraining_dataset


@click.command()
@click.option(
    "-b", "--batch_size", default=8, help="Batch size for training (per device)", type=int
)
@click.option("--fp16", default=True, help="Whether to use fp16 precision or not", type=bool)
@click.option("--lr", default=1e-4, help="Learning rate", type=float)
@click.option("--dataloader_num_workers", default=8, help="Number of dataloader workers", type=int)
@click.option("--deepspeed", default=False, help="Whether to use deepspeed or not", type=bool)
@click.option(
    "--gradient_accumulation_steps",
    default=1,
    help="Number of gradient accumulation steps",
    type=int,
)
@click.option(
    "--logging_dir", default=MLFLOW_TRACKING_URI, help="Mlflow logging/tracking URI", type=str
)
@click.option("--logging_steps", default=100, help="Logging interval", type=int)
@click.option("-m", "--max_steps", default=200, help="Number of training steps", type=int)
@click.option(
    "--overwrite_output_dir",
    default=False,
    help="Whether to override the output dir or not",
    type=bool,
)
@click.option(
    "--pretraining_file",
    default=PRETRAINING_PREPROCESSED_POSITIVE_DF_PATH,
    help="File used in pretraining containing the preprocessed training examples",
    type=str,
)
@click.option(
    "--save_limit", default=5, help="Maximum number of saved models/checkpoints", type=int
)
@click.option("--save_steps", default=5000, help="Checkpointing interval", type=int)
@click.option(
    "--training_dir",
    default=STONKGS_PRETRAINING_NO_NSP_DIR,
    help="Whether to override the output dir",
    type=str,
)
def pretrain_stonkgs(
    batch_size: int = 8,
    deepspeed: bool = False,
    fp16: bool = True,
    lr: float = 1e-4,
    dataloader_num_workers: int = 8,  # empirically determined value, I'm open to changing it :)
    gradient_accumulation_steps: int = 1,
    logging_dir: Optional[str] = MLFLOW_TRACKING_URI,
    logging_steps: int = 100,
    max_steps: int = 10000,
    overwrite_output_dir: bool = False,
    pretraining_file: str = PRETRAINING_PREPROCESSED_POSITIVE_DF_PATH,
    save_limit: int = 5,
    save_steps: int = 5000,
    training_dir: str = STONKGS_PRETRAINING_NO_NSP_DIR,
):
    """Run the pre-training procedure for the STonKGs model based on the transformers Trainer and TrainingArguments."""
    # Part of this code is taken from
    # https://github.com/huggingface/transformers/blob/master/examples/language-modeling/run_mlm.py

    # Logging with mlflow
    # End previous run
    mlflow.end_run()
    # Initialize mlflow run, set tracking URI to use the same experiment for all runs,
    # so that one can compare them
    mlflow.set_tracking_uri(logging_dir)
    mlflow.set_experiment("STonKGs Pre-Training (No NSP Ablation)")

    # Initialize the STonKGs model
    # config = None just fills up the required argument for automated method calls such as .from_pretrained, it will be
    # overridden anyways in __init__
    stonkgs_model = STonKGsForPreTraining(config=None, nlp_model_type=NLP_MODEL_TYPE)

    # Add the huggingface accelerator
    accelerator = Accelerator()
    # Use the device given by the `accelerator` object and put the model on there
    device = accelerator.device
    stonkgs_model.to(device)

    # Initialize the dataset
    pretraining_data = _load_pre_training_data(pretraining_preprocessed_path=pretraining_file)

    # Print logger statements about the allocated number of bytes of the training dataset
    logger.info(
        f"The number of bytes allocated by the dataset on the drive is {pretraining_data.dataset_size}"
    )
    logger.info(
        f"For comparison, here is the number of bytes allocated in memory by the training dataset"
        f": {total_allocated_bytes()}"
    )

    # Clean up cache files to be sure
    pretraining_data.cleanup_cache_files()

    # Accelerate the model
    stonkgs_model = accelerator.prepare(stonkgs_model)

    # Initialize the TrainingArguments
    training_args = TrainingArguments(
        output_dir=training_dir,
        overwrite_output_dir=overwrite_output_dir,
        # Use deepspeed with a specified config file for speedup
        deepspeed=DEEPSPEED_CONFIG_PATH if deepspeed else None,
        do_train=True,
        # Use fp16 to save space
        fp16=fp16,
        fp16_full_eval=fp16,
        per_device_train_batch_size=batch_size,
        max_steps=max_steps,  # Use max_steps rather than num_training_epochs
        learning_rate=lr,  # Default is to use that lr with a linear scheduler
        logging_dir=logging_dir,
        logging_steps=logging_steps,
        save_steps=save_steps,
        save_total_limit=save_limit,
        report_to=["mlflow"],
        # Make the dataloader faster by using pinning the memory and using multiple workers
        dataloader_pin_memory=True,
        dataloader_num_workers=dataloader_num_workers,
        # Effectively increase the batch size by gradient acc: batch_size = old_batch_size x grad_acc_steps
        gradient_accumulation_steps=gradient_accumulation_steps,
    )

    # Detecting last checkpoint
    last_checkpoint = None
    if (
        os.path.isdir(training_args.output_dir)
        and training_args.do_train
        and not training_args.overwrite_output_dir
    ):
        last_checkpoint = get_last_checkpoint(training_args.output_dir)
        if last_checkpoint is None and len(os.listdir(training_args.output_dir)) > 0:
            raise ValueError(
                f"Output directory ({training_args.output_dir}) already exists and is not empty. ",
                "Use --overwrite_output_dir to overcome.",
            )
        elif last_checkpoint is not None:
            logger.info(
                f"Checkpoint detected, resuming training at {last_checkpoint}. To avoid this behavior, change ",
                "the `--output_dir` or add `--overwrite_output_dir` to train from scratch.",
            )

    # Initialize the Trainer
    trainer = Trainer(
        model=stonkgs_model,
        args=training_args,
        train_dataset=pretraining_data,
    )
    # Delete the pretraining dataset from memory
    del pretraining_data
    # And train STonKGs to the moon
    train_result = trainer.train(resume_from_checkpoint=last_checkpoint)
    trainer.save_model()  # Saves the tokenizer too for easy upload
    metrics = train_result.metrics

    # Log all metrics
    trainer.log_metrics("train", metrics)
    trainer.save_metrics("train", metrics)
    trainer.save_state()


if __name__ == "__main__":
    # Run the pre-training procedure, overwrite the output dir for now (since we're only working with dummy data)
    # Distinguish between local and cluster execution by using different values (steps, batch size etc.) in the
    # CLI arguments
    pretrain_stonkgs()

    # (Optional) examine the runtime
    # with profiler.profile() as prof:
    #    with profiler.record_function("model_inference"):
    #        pretrain_stonkgs(overwrite_output_dir=True, max_steps=3)

    # logger.info(prof.key_averages().table(sort_by="cpu_time_total"))  # or replace by something like gpu_time_total



# -*- coding: utf-8 -*-

"""Filter the fine-tuning datasets by their majority classes."""

import logging
import os
from typing import Optional

import pandas as pd

from stonkgs.constants import (
    CELL_LINE_DIR,
    DISEASE_DIR,
    LOCATION_DIR,
    SPECIES_DIR,
)

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def filter_for_majority_classes(
    df: pd.DataFrame,
    n_classes: int = 5,
    name: str = "",
    output_path: Optional[str] = None,
) -> pd.DataFrame:
    """Filter out data entries that occur infrequently, return the dataframe with only n_class many majority classes."""
    # Remove the "-1" label (str) (since it can't be mapped to any ontology)
    df = df[df["class"] != "-1"]

    # Manually merge EFO:0000887 into UBERON:0002107, since it's deprecated
    df.replace("0000887", "0002107", inplace=True)

    # Remove any classes that are not in the n_classes most common classes
    # (value counts returns classes in descending order)
    labels_to_remove = df["class"].value_counts()[n_classes:].to_dict()
    labels_to_keep = df["class"].value_counts()[:n_classes].to_dict()

    # Some statistics
    # logger.info(f'labels removed due to low occurrence {labels_to_remove}')
    logger.info(f"{name} majority class occurrences {labels_to_keep}")
    logger.info(f"{name} raw triples {df.shape[0]}")

    # Remove all the entries that are labelled with a class that should be removed
    df = df[~df["class"].isin(list(labels_to_remove.keys()))]

    # Final length of the filtered df
    logger.info(f"{name} triples after filtering for {n_classes} classes: {df.shape[0]} \n")

    # Optional: Save the filtered df
    if output_path and len(name) > 0:
        df.to_csv(
            os.path.join(output_path, name + "_filtered_more_classes.tsv"), sep="\t", index=None
        )

    return df


if __name__ == "__main__":
    # Load all the unfiltered dataframes
    cell_line_df = pd.read_csv(os.path.join(CELL_LINE_DIR, "cell_line.tsv"), sep="\t")
    disease_df = pd.read_csv(os.path.join(DISEASE_DIR, "disease.tsv"), sep="\t")
    location_df = pd.read_csv(os.path.join(LOCATION_DIR, "location.tsv"), sep="\t")
    species_df = pd.read_csv(os.path.join(SPECIES_DIR, "species.tsv"), sep="\t")

    # See how many entries there are for each filtered dataframe
    filter_for_majority_classes(
        cell_line_df, name="cell_line", n_classes=10, output_path=CELL_LINE_DIR
    )
    filter_for_majority_classes(disease_df, name="disease", n_classes=10, output_path=DISEASE_DIR)
    filter_for_majority_classes(location_df, name="location", n_classes=5, output_path=LOCATION_DIR)
    filter_for_majority_classes(species_df, name="species", n_classes=3, output_path=SPECIES_DIR)



# -*- coding: utf-8 -*-

"""Command line interface for :mod:`stonkgs`.

Why does this file exist, and why not put this in ``__main__``? You might be tempted to import things from ``__main__``
later, but that will cause problems--the code will get executed twice:

- When you run ``python3 -m stonkgs`` python will execute``__main__.py`` as a script.
  That means there won't be any ``stonkgs.__main__`` in ``sys.modules``.
- When you import __main__ it will get executed again (as a module) because
  there's no ``stonkgs.__main__`` in ``sys.modules``.

.. seealso:: https://click.palletsprojects.com/en/7.x/setuptools/#setuptools-integration
"""

import logging

import click  # type: ignore

__all__ = ["main"]

logger = logging.getLogger(__name__)


@click.group()
@click.version_option()
def main():
    """CLI for stonkgs."""
    click.echo("Please check the documentation to learn how to train the STonKGs model")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""STonKGs model architecture components."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import lru_cache
from typing import Optional

import torch
from torch import nn
from transformers import (
    BertConfig,
    BertForPreTraining,
    BertModel,
    BertTokenizer,
)
from transformers.models.bert.modeling_bert import BertForPreTrainingOutput, BertLMPredictionHead

from stonkgs.constants import EMBEDDINGS_PATH, NLP_MODEL_TYPE
from stonkgs.models.kg_baseline_model import prepare_df

# Initialize logger
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


@dataclass
class BertForPreTrainingOutputWithPooling(BertForPreTrainingOutput):
    """Overriding the BertForPreTrainingOutput class to further include the pooled output."""

    pooler_output: Optional[torch.FloatTensor] = None


class STonKGsELMPredictionHead(BertLMPredictionHead):
    """Custom masked entity and language modeling (ELM) head used to predict both entities and text tokens."""

    def __init__(self, config):
        """Initialize the ELM head based on the (hyper)parameters in the provided BertConfig."""
        super().__init__(config)

        # There are two different "decoders": The first half of the sequence is projected onto the dimension of
        # the text vocabulary index, the second half is projected onto the dimension of the kg vocabulary index
        # 1. Text decoder
        self.text_decoder = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        # 2. Entity decoder
        self.entity_decoder = nn.Linear(config.hidden_size, config.kg_vocab_size, bias=False)

        # Determine half of the maximum sequence length based on the config
        self.half_length = config.max_position_embeddings // 2

        # Set the biases differently for the decoder layers
        self.text_bias = nn.Parameter(torch.zeros(config.vocab_size))
        self.entity_bias = nn.Parameter(torch.zeros(config.kg_vocab_size))

        # Need a link between the two variables so that the bias is correctly resized with `resize_token_embeddings`
        self.decoder.text_bias = self.text_bias
        self.decoder.entity_bias = self.entity_bias

    def forward(self, hidden_states):
        """Map hidden states to values for the text vocab (first half) and kg vocab (second half)."""
        # Common transformations (dense layer, layer norm + activation function) performed on both text and entities
        # transform is initialized in the parent BertLMPredictionHead class
        hidden_states = self.transform(hidden_states)

        # The first half is processed with the text decoder, the second with the entity decoder to map to the text
        # vocab size and kg vocab size, respectively
        text_hidden_states_to_vocab = self.text_decoder(hidden_states[:, : self.half_length])
        ent_hidden_states_to_kg_vocab = self.entity_decoder(hidden_states[:, self.half_length :])

        return text_hidden_states_to_vocab, ent_hidden_states_to_kg_vocab


class STonKGsForPreTraining(BertForPreTraining):
    """Create the pre-training part of the STonKGs model based on both text and entity embeddings."""

    def __init__(
        self,
        config,  # Required for automated methods such as .from_pretrained in classes that inherit from this one,
        # the config is loaded from scratch later on anyways
        nlp_model_type: str = NLP_MODEL_TYPE,
        kg_embedding_dict_path: str = EMBEDDINGS_PATH,
    ):
        """Initialize the model architecture components of STonKGs."""
        # Initialize the KG dict from the file here, rather than passing it as a parameter, so that it can
        # be loaded from a checkpoint
        kg_embedding_dict = prepare_df(kg_embedding_dict_path)

        # Add the number of KG entities to the default config of a standard BERT model
        config = BertConfig.from_pretrained(nlp_model_type)
        config.update({"kg_vocab_size": len(kg_embedding_dict)})
        # Initialize the underlying BertForPreTraining model that will be used to build the STonKGs Transformer layers
        super().__init__(config)

        # Override the standard MLM head: In the underlying BertForPreTraining model, change the MLM head to the custom
        # STonKGsELMPredictionHead so that it can be used on the concatenated text/entity input
        self.cls.predictions = STonKGsELMPredictionHead(config)

        # Language Model (LM) backbone initialization (pre-trained BERT to get the initial embeddings)
        # based on the specified nlp_model_type (e.g. BioBERT)
        self.lm_backbone = BertModel.from_pretrained(nlp_model_type)
        # Put the LM backbone on the GPU if possible
        if torch.cuda.is_available():
            self.lm_backbone.to("cuda")
        # Freeze the parameters of the LM backbone so that they're not updated during training
        # (We only want to train the STonKGs Transformer layers)
        for param in self.lm_backbone.parameters():
            param.requires_grad = False
        # Get the separator, mask and unknown token ids from a nlp_model_type specific tokenizer
        self.lm_sep_id = BertTokenizer.from_pretrained(nlp_model_type).sep_token_id  # usually 102
        self.lm_mask_id = BertTokenizer.from_pretrained(nlp_model_type).mask_token_id  # usually 103
        self.lm_unk_id = BertTokenizer.from_pretrained(nlp_model_type).unk_token_id  # usually 100

        # KG backbone initialization
        # Get numeric indices for the KG embedding vectors except for the sep, unk, mask ids which are reserved for the
        # LM [SEP] embedding vectors (see below)
        numeric_indices = list(range(len(kg_embedding_dict) + 3))
        # Keep the numeric indices of the special tokens free, don't put the kg embeds there
        for special_token_id in [self.lm_sep_id, self.lm_mask_id, self.lm_unk_id]:
            numeric_indices.remove(special_token_id)

        # Generate numeric indices for the KG node names (iterating .keys() is deterministic)
        self.kg_idx_to_name = {i: key for i, key in zip(numeric_indices, kg_embedding_dict.keys())}
        # Initialize KG index to embeddings based on the provided kg_embedding_dict
        self.kg_backbone = {
            i: torch.tensor(kg_embedding_dict[self.kg_idx_to_name[i]]).to(self.lm_backbone.device)
            for i in self.kg_idx_to_name.keys()
        }
        # Add the MASK, SEP and UNK (LM backbone) embedding vectors to the KG backbone so that the labels are correctly
        # identified in the loss function later on
        # [0][0][0] is required to get the shape from batch x seq_len x hidden_size to hidden_size
        for special_token_id in [self.lm_sep_id, self.lm_mask_id, self.lm_unk_id]:
            self.kg_backbone[special_token_id] = self.lm_backbone(
                torch.tensor([[special_token_id]]).to(self.lm_backbone.device),
            )[0][0][0]

    @classmethod
    @lru_cache(maxsize=32)
    def from_default_pretrained(cls, **kwargs) -> STonKGsForPreTraining:
        """Get the default pre-trained STonKGs model."""
        return cls.from_pretrained("stonkgs/stonkgs-150k", **kwargs)

    def forward(
        self,
        # required parameters
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        masked_lm_labels=None,
        ent_masked_lm_labels=None,
        next_sentence_labels=None,
        # determined the return type
        return_dict=None,
        # in case certain masks do need to be canceled out
        head_mask=None,
    ):
        """Perform one forward pass for a given sequence of text_input_ids + ent_input_ids."""
        # The code is based on CoLAKE: https://github.com/txsun1997/CoLAKE/blob/master/pretrain/model.py

        # Use the LM backbone to get the pre-trained token embeddings
        # batch x half_length x hidden_size
        # The first element of the returned tuple from the LM backbone forward() pass is the sequence of hidden states
        token_embeddings = self.lm_backbone(input_ids[:, : self.cls.predictions.half_length])[0]

        # Use the KG backbone to obtain the pre-trained entity embeddings
        # batch x half_length x hidden_size
        ent_embeddings = torch.stack(
            [
                # for each numeric index in the random walks sequence: get the embedding vector from the KG backbone
                torch.stack([self.kg_backbone[i.item()] for i in j])
                # for each example in the batch: get the random walks sequence
                for j in input_ids[:, self.cls.predictions.half_length :]
            ],
        )

        # Concatenate token and entity embeddings obtained from the LM and KG backbones and cast to float
        # batch x seq_len x hidden_size
        inputs_embeds = (
            torch.cat(
                [token_embeddings, ent_embeddings.to(token_embeddings.device)],
                dim=1,
            )
            .type(torch.FloatTensor)
            .to(self.device)
        )

        # Get the hidden states from the basic STonKGs Transformer layers
        # batch x half_length x hidden_size
        outputs = self.bert(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            head_mask=head_mask,
            return_dict=None,
        )
        # batch x seq_len x hidden_size
        sequence_output, pooled_output = outputs[:2]

        # Generate the prediction scores (mapping to text and entity vocab sizes + NSP) for the training objectives
        # Seq_relationship_score = NSP score
        # prediction_scores = Text MLM and Entity "MLM" scores
        prediction_scores, seq_relationship_score = self.cls(sequence_output, pooled_output)
        # The custom STonKGsELMPredictionHead returns a pair of prediction score sequences for tokens and entities,
        # respectively
        token_prediction_scores, entity_predictions_scores = prediction_scores

        # Calculate the loss
        total_loss = None
        if (
            masked_lm_labels is not None
            and ent_masked_lm_labels is not None
            and next_sentence_labels is not None
        ):
            loss_fct = nn.CrossEntropyLoss()
            # 1. Text-based MLM
            masked_lm_loss = loss_fct(
                token_prediction_scores.view(-1, self.config.vocab_size),
                masked_lm_labels.view(-1),
            )
            # 2. Entity-based masked "language" (entity) modeling
            ent_masked_lm_loss = loss_fct(
                entity_predictions_scores.view(-1, self.config.kg_vocab_size),
                ent_masked_lm_labels.view(-1),
            )
            # 3. Next "sentence" loss: Whether a text and random walk sequence belong together or not
            # next_sentence_loss = loss_fct(
            #     seq_relationship_score.view(-1, 2),
            #    next_sentence_labels.view(-1),
            # )
            # Total loss = the sum of the individual training objective losses
            # !! Leave out NSP loss in the ablation !!
            total_loss = masked_lm_loss + ent_masked_lm_loss  # + next_sentence_loss

        if not return_dict:
            output = (prediction_scores, seq_relationship_score) + outputs[2:]
            return ((total_loss,) + output) if total_loss is not None else output

        return BertForPreTrainingOutputWithPooling(
            loss=total_loss,
            prediction_logits=prediction_scores,
            seq_relationship_logits=seq_relationship_score,
            hidden_states=sequence_output,
            attentions=outputs.attentions,
            pooler_output=pooled_output,
        )



# -*- coding: utf-8 -*-

"""Filter out duplicate text evidences (the ones that appear more than two times)."""

import logging
import re
import os
from collections import Counter
from typing import Optional

import pandas as pd
from sklearn.model_selection import train_test_split
from transformers import BertTokenizer

from stonkgs.constants import (
    CELL_LINE_DIR,
    DISEASE_DIR,
    EMBEDDINGS_PATH,
    LOCATION_DIR,
    NLP_MODEL_TYPE,
    SPECIES_DIR,
    RELATION_TYPE_DIR,
)
from stonkgs.data.indra_for_pretraining import prepare_df

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def filter_out_duplicates(
    df: pd.DataFrame,
    name: str = "",
) -> pd.DataFrame:
    """Filter for unique text evidences in the data entries, avoiding repeating evidences."""

    # Record the length before for logging purposes
    len_before = len(df)

    # Filter for unique evidences
    df = df.drop_duplicates(subset="evidence")

    # Record the length after and log it
    len_after = len(df)
    logger.info(
        f"{name}: {len_before} (before), {len_after} (after), {len_before-len_after} (# removed)"
    )

    return df


def apply_kg_filtering(
    df: pd.DataFrame,
    embedding_name_to_vector_path: str = EMBEDDINGS_PATH,
    name: str = "",
) -> pd.DataFrame:
    """Filters out entries in the fine-tuning dataset that contain nodes which are not part of the pre-trained KG."""
    kg_embed_dict = prepare_df(embedding_name_to_vector_path)
    original_length = len(df)
    df = df[
        df["source"].isin(kg_embed_dict.keys()) & df["target"].isin(kg_embed_dict.keys())
    ].reset_index(drop=True)
    new_length = len(df)
    logger.info(
        f"For {name}, {original_length - new_length} out of {original_length} triples are left out because "
        f"they contain nodes which are not present in the pre-training data"
    )
    return df


def reduce_dataset_size(
    df: pd.DataFrame,
    max_dataset_size: int = 10000,
    class_name: str = "class",
    random_seed: int = 42,
) -> pd.DataFrame:
    """Creates a stratified subset of the original dataset if it exceeds max_dataset_size."""
    # Cut the dataset down to max_dataset_size (deterministically!) using StratifiedShuffleSplit if needed:
    # (this is not an actual train/test split, this is just for getting a dataset of size max_dataset_size in a
    # stratified and deterministic manner)
    if max_dataset_size < len(df):
        if class_name == "class":
            df = train_test_split(
                df,
                train_size=max_dataset_size,
                random_state=random_seed,
                stratify=df[class_name],
            )[0]
        else:
            df = train_test_split(
                df,
                train_size=max_dataset_size * 2,
                random_state=random_seed,
                stratify=df["interaction"],
            )[0]

            df = train_test_split(
                df,
                train_size=max_dataset_size,
                random_state=random_seed,
                stratify=df["polarity"],
            )[0]

            # Just some extra info on the relation type distribution afterwards
            if name == "relation_type":
                logger.info(f"Polarity: {Counter(df['polarity'])}")
                logger.info(f"Interaction: {Counter(df['interaction'])}")

    return df


def filter_out_special_character_sequences(
    df: pd.DataFrame,
    min_tokens: int = 50,
    evidence_col_name: str = "evidence",
    name: str = "",
) -> pd.DataFrame:
    """Filter out special character sequences."""
    counter = 0
    idx_to_remove = []
    initial_length = len(df)
    # filtering_pattern = re.compile("((\\\\)+u([\w]+))|(\[[^\s]+\])|(XREF)")
    tokenizer = BertTokenizer.from_pretrained(NLP_MODEL_TYPE)

    for idx, row in df.iterrows():
        if len(tokenizer.tokenize(row[evidence_col_name])) < min_tokens:
            idx_to_remove.append(idx)
        elif any([x in row[evidence_col_name] for x in ["[", "]", "XREF", "\\u"]]):
            counter += 1
            row[evidence_col_name] = row[evidence_col_name].replace("[", "")
            row[evidence_col_name] = row[evidence_col_name].replace("]", "")
            row[evidence_col_name] = row[evidence_col_name].replace(r"\\u", "")
            row[evidence_col_name] = row[evidence_col_name].replace("XREF", "")
            df.iloc[idx] = row

    df.drop(index=idx_to_remove, inplace=True)
    df.reset_index(inplace=True, drop=True)

    logger.info(
        f"For {name}, {counter} out of {initial_length} many entries contained the specified special characters"
        f" and {len(idx_to_remove)} many entries were removed because they are too short, resulting in "
        f"{len(df)} many entries"
    )

    return df


if __name__ == "__main__":
    all_names = [
        "cell_line",
        "disease",
        "location",
        "species",
        "relation_type",
    ]
    all_dirs = [
        CELL_LINE_DIR,
        DISEASE_DIR,
        LOCATION_DIR,
        SPECIES_DIR,
        RELATION_TYPE_DIR,
    ]

    for name, directory in zip(all_names, all_dirs):
        # Load the unfiltered dataframe
        if name == "relation_type":
            task_specific_df = pd.read_csv(os.path.join(directory, name + ".tsv"), sep="\t")
        else:
            task_specific_df = pd.read_csv(
                os.path.join(directory, name + "_filtered_more_classes.tsv"), sep="\t"
            )

        # 1. Remove all entries with nodes that are not in the KG
        task_specific_df = apply_kg_filtering(task_specific_df, name=name)
        # 2. Remove special characters in the evidences
        task_specific_df = filter_out_special_character_sequences(task_specific_df, name=name)
        # 3. Remove duplicates
        task_specific_df = filter_out_duplicates(task_specific_df, name=name)
        # 4. Reduce dataset size if needed
        # if name == "relation_type":
        #     task_specific_df = reduce_dataset_size(task_specific_df, class_name="interaction")
        # else:
        #     task_specific_df = reduce_dataset_size(task_specific_df, class_name="class")

        # Save the filtered df
        task_specific_df.to_csv(
            os.path.join(directory, name + "_more_classes_no_duplicates.tsv"), sep="\t", index=None
        )
